import re
import os
import time
import glob
import copy
from datetime import datetime
from dateutil.parser import parse as dtparse
import json
import asyncio
from functools import partial
import subprocess

from biothings.utils.mongo import get_previous_collection, get_target_db
from biothings.utils.hub_db import get_src_build, get_source_fullname
import biothings.utils.aws as aws
from biothings.utils.dataload import update_dict_recur
from biothings.utils.jsondiff import make as jsondiff
from biothings.utils.loggers import get_logger
from biothings.utils.manager import BaseManager, BaseStatusRegisterer
from biothings.utils.backend import DocMongoBackend
from biothings import config as btconfig
from biothings.utils.hub import publish_data_version, template_out
from biothings.hub.databuild.backend import generate_folder, create_backend
from biothings.hub import RELEASEMANAGER_CATEGORY, RELEASER_CATEGORY
from biothings.hub.datarelease.releasenote import ReleaseNoteTxt
from biothings.hub.datarelease import set_pending_to_publish

# default from config
logging = btconfig.logger


class PublisherException(Exception):
    pass


class BasePublisher(BaseManager, BaseStatusRegisterer):
    def __init__(self, envconf, log_folder, es_backups_folder, *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.envconf = envconf
        self.log_folder = log_folder
        self.es_backups_folder = es_backups_folder
        self.ti = time.time()
        self.setup()

    def clean_stale_status(self):
        src_build = get_src_build()
        for build in src_build.find():
            dirty = False
            for job in build.get("jobs", []):
                if job.get("status", "").endswith("publishing"):
                    logging.warning(
                        "Found stale build '%s', marking publish status as 'canceled'"
                        % build["_id"])
                    job["status"] = "canceled"
                    dirty = True
            if dirty:
                src_build.replace_one({"_id": build["_id"]}, build)

    @property
    def category(self):
        return RELEASER_CATEGORY

    @property
    def collection(self):
        return get_src_build()

    def setup(self):
        self.setup_log()

    def setup_log(self):
        self.logger, self.logfile = get_logger(self.category, self.log_folder)

    def get_predicates(self):
        return []

    def get_pinfo(self):
        """
        Return dict containing information about the current process
        (used to report in the hub)
        """
        pinfo = {
            "category": self.category,
            "source": "",
            "step": "",
            "description": ""
        }
        preds = self.get_predicates()
        if preds:
            pinfo["__predicates__"] = preds
        return pinfo

    def register_status(self,
                        bdoc,
                        status,
                        transient=False,
                        init=False,
                        **extra):
        BaseStatusRegisterer.register_status(self,
                                             bdoc,
                                             "publish",
                                             status,
                                             transient=transient,
                                             init=init,
                                             **extra)

    def load_build(self, key_name, stage=None):
        if stage is None:
            # picke build doc searching for snapshot then diff key if not found
            return self.load_doc(key_name, "snapshot") or self.load_doc(
                key_name, "diff")
        else:
            return self.load_doc(key_name, stage)

    def template_out_conf(self, build_doc):
        # don't bother exploring certain keys, convert the whole a string
        # and replace whatever we find
        try:
            strconf = template_out(json.dumps(self.envconf), build_doc)
            return json.loads(strconf)
        except Exception as e:
            self.logger.exception("Coudn't template out configuration: %s" % e)
            raise

    def create_bucket(self, bucket_conf, credentials):
        aws_key = credentials.get("access_key")
        aws_secret = credentials.get("secret_key")
        # create bucket if needed
        aws.create_bucket(name=bucket_conf["bucket"],
                          region=bucket_conf["region"],
                          aws_key=aws_key,
                          aws_secret=aws_secret,
                          acl=bucket_conf.get("acl"),
                          ignore_already_exists=True)
        if bucket_conf.get("website"):
            aws.set_static_website(bucket_conf["bucket"],
                                   aws_key=aws_key,
                                   aws_secret=aws_secret)

    def trigger_release_note(self, doc, **kwargs):
        """
        Launch a release note generation given a src_build document. In order to
        know the first collection to compare with, get_previous_collection()
        method is used. release_note() method will get **kwargs for more optional
        parameters.
        """
        new_db_col_names = doc["_id"]
        old_db_col_names = get_previous_collection(new_db_col_names)
        self.release_note(old_db_col_names, new_db_col_names, **kwargs)

    def get_pre_post_previous_result(self, build_doc, key_value):
        """
        In order to start a pre- or post- pipeline, a first previous result, fed
        all along the pipeline to the next step, has to be defined, and depends
        on the type of publisher.
        """
        raise NotImplementedError("implement me in sub-class")

    def run_pre_post(self, key, stage, key_value, repo_conf, build_doc):
        """
        Run pre- and post- publish steps (stage) for given key (eg. "snapshot", "diff").
        key_value is the value of the key inside "key" dict (such as a snapshot name or a build name)
        These steps are defined in config file.
        """
        # determine previous result as the starting point of the pipeline,
        # depending on the type of publishing (key)
        previous_result = self.get_pre_post_previous_result(
            build_doc, key_value)
        steps = repo_conf.get("publish", {}).get(stage, {}).get(
            key, [])  # publish[pre|post][snapshot|diff]
        assert isinstance(
            steps,
            list), "'%s' stage must be a list, got: %s" % (stage, repr(steps))
        action_done = []
        for step_conf in steps:
            try:
                action = step_conf["action"]
                self.logger.info("Processing stage '%s-%s-publish': %s" %
                                 (stage, key, action))
                # first try user-defined methods
                # it can be named (first to match is picked):
                # see list below
                found = False
                for tpl in [
                        "step_%(stage)s_publish_%(action)s",
                        "step_publish_%(action)s", "step_%(stage)s_%(action)s"
                ]:
                    methname = tpl % {"stage": stage, "action": action}
                    if hasattr(self, methname):
                        found = True
                        previous_result = getattr(self,
                                                  methname)(step_conf,
                                                            build_doc,
                                                            previous_result)
                        break
                if not found:
                    # default to generic one
                    previous_result = getattr(self, "step_%s" % action)(
                        step_conf, build_doc, previous_result)

                action_done.append({"name": action, "result": previous_result})

            except AttributeError:
                raise ValueError("No such %s-%s-publish step '%s'" %
                                 (stage, key, action))

        return action_done

    def step_archive(self, step_conf, build_doc, previous):
        archive_name = step_conf["name"]
        archive_file = os.path.join(self.es_backups_folder, archive_name)
        if step_conf["format"] == "tar.xz":
            # -J is for "xz"
            tarcmd = [
                "tar",
                "cJf",
                archive_file,  # could be replaced if "split"
                "-C",
                self.es_backups_folder,
                previous["settings"]["location"],
            ]
            if step_conf.get("split"):
                part = "%s.part." % archive_file
                tarcmd[2] = "-"
                tarps = subprocess.Popen(tarcmd, stdout=subprocess.PIPE)
                splitcmd = ["split", "-", "-b", step_conf["split"], "-d", part]
                ps = subprocess.Popen(splitcmd, stdin=tarps.stdin)
                ret_code = ps.wait()
                if ret_code != 0:
                    raise PublisherException("Archiving failed, code: %s" %
                                             ret_code)
                else:
                    flist = glob.glob("%s.*" % part)
                    if len(flist) == 1:
                        # no even split, go back to single archive file
                        os.rename(flist[0], archive_file)
                        self.logger.info("Tried to split archive, but only one part was produced,"
                                         + "returning single archive file: %s" % archive_file)
                        return archive_file
                    else:
                        # produce a json file with metadata about the splits
                        jsonfile = "%s.json" % outfile
                        json.dump({
                            "filename": outfile,
                            "parts": flist
                        }, open(jsonfile, "w"))
                        self.logger.info(
                            "Archive split into %d parts, metadata stored in: %s"
                            % (len(flist), jsonfile))
                        return jsonfile
            else:
                subprocess.check_output(tarcmd)
                self.logger.info("Archive: %s" % archive_file)
                return archive_file
        else:
            raise ValueError("Only 'tar.xz' format supported for now, got %s" %
                             repr(step_conf["format"]))

    def step_upload(self, step_conf, build_doc, previous):
        if step_conf["type"] == "s3":
            return self.step_upload_s3(step_conf, build_doc, previous)
        else:
            raise ValueError(
                "Only 's3' upload type supported for now, got %s" %
                repr(step_conf["type"]))

    def step_upload_s3(self, step_conf, build_doc, previous):
        aws_key = self.envconf.get("cloud", {}).get("access_key")
        aws_secret = self.envconf.get("cloud", {}).get("secret_key")
        # create bucket if needed
        self.create_bucket(bucket_conf=step_conf,
                           credentials=self.envconf.get("cloud", {}))
        if step_conf.get("file"):
            basename = step_conf["file"]
            uploadfunc = aws.send_s3_big_file
        elif step_conf.get("folder"):
            basename = step_conf["folder"]
            uploadfunc = aws.send_s3_folder
        else:
            raise ValueError(
                "Can't find 'file' or 'folder' key, don't know what to upload")
        archive_path = os.path.join(self.es_backups_folder, basename)
        self.logger.info("Uploading: %s" % archive_path)
        uploadfunc(archive_path,
                   os.path.join(step_conf["base_path"], basename),
                   overwrite=step_conf.get("overwrite", False),
                   aws_key=aws_key,
                   aws_secret=aws_secret,
                   s3_bucket=step_conf["bucket"])
        return {
            "type": "s3",
            "key": basename,
            "base_path": step_conf["base_path"],
            "bucket": step_conf["bucket"]
        }

    def get_release_note_filename(self, build_version):
        return "release_%s" % build_version

    def publish_release_notes(self,
                              release_folder,
                              build_version,
                              s3_release_folder,
                              s3_release_bucket,
                              aws_key,
                              aws_secret,
                              prefix="release_"):
        release_note = self.get_release_note_filename(build_version)
        s3basedir = os.path.join(s3_release_folder, build_version)
        notes = glob.glob(os.path.join(release_folder, "%s.*" % release_note))
        self.logger.info(
            "Uploading release notes from '%s' to s3 folder '%s'" %
            (notes, s3basedir))
        for note in notes:
            if os.path.exists(note):
                s3key = os.path.join(s3basedir, os.path.basename(note))
                aws.send_s3_file(
                    note,
                    s3key,
                    aws_key=self.envconf.get("cloud", {}).get("access_key"),
                    aws_secret=self.envconf.get("cloud", {}).get("secret_key"),
                    s3_bucket=s3_release_bucket,
                    overwrite=True)
        # specify release note URLs in metadata
        rel_txt_url = aws.get_s3_url(
            os.path.join(s3basedir, "%s.txt" % release_note),
            aws_key=self.envconf.get("cloud", {}).get("access_key"),
            aws_secret=self.envconf.get("cloud", {}).get("secret_key"),
            s3_bucket=s3_release_bucket)
        rel_json_url = aws.get_s3_url(
            os.path.join(s3basedir, "%s.json" % release_note),
            aws_key=self.envconf.get("cloud", {}).get("access_key"),
            aws_secret=self.envconf.get("cloud", {}).get("secret_key"),
            s3_bucket=s3_release_bucket)
        urls = {}
        if rel_txt_url:
            urls["txt"] = {"url": rel_txt_url}
        if rel_json_url:
            urls["json"] = {"url": rel_json_url}

        return urls


class SnapshotPublisher(BasePublisher):
    def __init__(self, snapshot_manager, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.snapshot_manager = snapshot_manager
        self.setup()

    def get_pre_post_previous_result(self, build_doc, key_value):
        assert len(build_doc["snapshot"][key_value]["repository"]) == 1
        repo_name = list(
            build_doc["snapshot"][key_value]["repository"].keys())[0]
        previous_result = build_doc["snapshot"][key_value]["repository"][
            repo_name]
        return previous_result

    def run_pre_publish_snapshot(self, snapshot_name, repo_conf, build_doc):
        return self.run_pre_post("snapshot", "pre", snapshot_name, repo_conf,
                                 build_doc)

    def publish(self,
                snapshot,
                build_name=None,
                previous_build=None,
                steps=["pre", "meta", "post"]):
        """
        Publish snapshot metadata to S3. If snapshot repository is of type "s3", data isn't actually
        uploaded/published since it's already there on s3. If type "fs", some "pre" steps can be added
        to the RELEASE_CONFIG paramater to archive and upload it to s3. Metadata about the snapshot,
        release note, etc... is then uploaded in correct buckets as defined in config, and "post"
        steps can be run afterward.

        Though snapshots don't need any previous version to be applied on, a release note with significant changes
        between current snapshot and a previous version could have been generated. By default, snapshot name is
        used to pick one single build document and from the document, get the release note information.
        """
        try:
            if build_name:
                bdoc = self.load_build(build_name)
            else:
                bdoc = self.load_build(snapshot, "snapshot")
        except Exception as e:
            self.exception(
                "Error loading build document using snapshot named '%s': %s" %
                (snapshot, e))
            raise
        if isinstance(bdoc, list):
            raise PublisherException(
                "Snapshot '%s' found in more than one builds: %s." %
                (snapshot, [d["_id"] for d in bdoc])
                + " Specify which one with 'build_name'")
        if type(steps) == str:
            steps = [steps]
        if not bdoc:
            raise PublisherException(
                "No build document found with a snapshot name '%s' associated to it"
                % snapshot)

        # instantiate publishing environment
        self.envconf = self.template_out_conf(bdoc)

        # check if a release note is associated to the build document
        release_folder = None
        if previous_build is None and "release_note" in bdoc:
            previous_build = list(bdoc["release_note"].keys())
            if len(previous_build) != 1:
                raise PublisherException("More than one release note found, "
                                         + "generated with following builds: %s" % previous_build)
            else:
                previous_build = previous_build.pop()

        assert previous_build, "Couldn't find previous build %s" % bdoc.keys()
        release_folder = generate_folder(btconfig.RELEASE_PATH, previous_build,
                                         bdoc["_id"])

        assert release_folder, "No release folder found, can't publish"

        s3_release_folder = self.envconf["release"]["folder"]
        s3_release_bucket = self.envconf["release"]["bucket"]
        self.create_bucket(bucket_conf=self.envconf["release"],
                           credentials=self.envconf.get("cloud", {}))

        # hold error/exception on each step
        got_error = None

        @asyncio.coroutine
        def do():
            jobs = []
            pinfo = self.get_pinfo()
            pinfo["step"] = "publish"
            pinfo["source"] = snapshot

            def done(f, step):
                try:
                    res = f.result()
                    self.register_status(bdoc,
                                         "success",
                                         job={
                                             "step": step,
                                             "result": res
                                         },
                                         publish={
                                             "full": {
                                                 snapshot: {
                                                     "conf": self.envconf,
                                                     step: res
                                                 }
                                             }
                                         })
                except Exception as e:
                    nonlocal got_error
                    got_error = e
                    self.register_status(bdoc,
                                         "failed",
                                         job={
                                             "step": step,
                                             "err": str(e)
                                         },
                                         publish={
                                             "full": {
                                                 snapshot: {
                                                     "conf": self.envconf,
                                                     step: {
                                                         "err": str(e)
                                                     }
                                                 }
                                             }
                                         })
                    self.logger.exception(
                        "Error while running pre-publish: %s" % got_error)

            if "_meta" not in bdoc:
                raise PublisherException(
                    "No metadata (_meta) found in build document")

            if "pre" in steps:
                # then we upload all the folder content
                pinfo["step"] = "pre"
                self.logger.info("Running pre-publish step")
                self.register_status(bdoc,
                                     "pre",
                                     transient=True,
                                     init=True,
                                     job={"step": "pre"},
                                     publish={"full": {
                                         snapshot: {}
                                     }})
                job = yield from self.job_manager.defer_to_thread(
                    pinfo,
                    partial(self.pre_publish, snapshot, self.envconf, bdoc))
                job.add_done_callback(partial(done, step="pre"))
                yield from job
                if got_error:
                    raise got_error
                jobs.append(job)

            if "meta" in steps:
                # TODO: this is a blocking call
                # snapshot at this point can be totally different than original
                # target_name but we still use it to potentially custom indexed
                # (anyway, it's just to access some snapshot info so default indexer
                # will work)
                build_version = bdoc["_meta"]["build_version"]
                self.logger.info(
                    "Generating JSON metadata for full release '%s'" %
                    build_version)
                # generate json metadata about this diff release
                assert snapshot, "Missing snapshot name information"
                if getattr(btconfig, "SKIP_CHECK_VERSIONS", None):
                    self.logger.info(
                        "SKIP_CHECK_VERSIONS %s, no version check will be performed on full metadata"
                        % repr(btconfig.SKIP_CHECK_VERSIONS))
                else:
                    assert getattr(
                        btconfig, "BIOTHINGS_VERSION", "master"
                    ) != "master", "I won't publish data refering BIOTHINGS_VERSION='master'"
                    assert getattr(
                        btconfig, "APP_VERSION", "master"
                    ) != "master", "I won't publish data refering APP_VERSION='master'"
                    assert getattr(btconfig, "STANDALONE_VERSION",
                                   None), "STANDALONE_VERSION not defined"
                full_meta = {
                    "type": "full",
                    "build_version": build_version,
                    "target_version": build_version,
                    "release_date": datetime.now().isoformat(),
                    "app_version": btconfig.APP_VERSION,
                    "biothings_version": btconfig.BIOTHINGS_VERSION,
                    "standalone_version": btconfig.STANDALONE_VERSION,
                    "metadata": {
                        "repository": bdoc["snapshot"][snapshot]["repository"],
                        "snapshot_name": snapshot,
                    }
                }
                # if snapshot tyoe is "fs" (so it means it's stored locally) and we publish (so it means we want it to
                # be available remotely) it means we should have an pre-"upload" step in the publish pipeline
                # let's try to get the archive url
                repo_name = list(
                    bdoc["snapshot"][snapshot]["repository"].keys())[0]
                if bdoc["snapshot"][snapshot]["repository"][repo_name][
                        "type"] == "fs":
                    pre_steps = bdoc.get("publish",
                                         {}).get("full",
                                                 {}).get(snapshot,
                                                         {}).get("pre", [])
                    try:
                        assert pre_steps, "No pre-steps found, expecting pre-upload step"
                        upload_step = [
                            step for step in pre_steps
                            if step["name"] == "upload"
                        ]
                        assert len(
                            upload_step
                        ) == 1, "Expecting one pre-upload step, got %s" % repr(
                            upload_step)
                        upload_step = upload_step.pop()
                        res = upload_step["result"]
                        assert res[
                            "type"] == "s3", "Only archived uploaded to S3 are currently supported"
                        url = aws.get_s3_url(
                            s3key=os.path.join(res["base_path"], res["key"]),
                            aws_key=self.envconf.get("cloud",
                                                     {}).get("access_key"),
                            aws_secret=self.envconf.get("cloud",
                                                        {}).get("secret_key"),
                            s3_bucket=res["bucket"])
                        full_meta["metadata"]["archive_url"] = url
                    except Exception as e:
                        raise PublisherException("Repository for snapshot '%s' is type 'fs' but " % snapshot
                                                 + "coudln't determine archive URL to publish: %s" % e)

                if release_folder:
                    if os.path.exists(release_folder):
                        try:
                            self.register_status(
                                bdoc,
                                "publishing",
                                transient=True,
                                init=True,
                                job={"step": "release-note"},
                                publish={"full": {
                                    snapshot: {}
                                }})
                            # ok, we have something in that folder, just pick the release note files
                            # (we can generate diff + snaphost at the same time, so there could be diff files in that folder
                            # from a diff process done before. release notes will be the same though)
                            urls = self.publish_release_notes(
                                release_folder,
                                build_version,
                                s3_release_folder,
                                s3_release_bucket,
                                aws_key=self.envconf.get("cloud",
                                                         {}).get("access_key"),
                                aws_secret=self.envconf.get(
                                    "cloud", {}).get("secret_key"))
                            full_meta.setdefault("changes", {})
                            full_meta["changes"].update(urls)
                            s3basedir = os.path.join(s3_release_folder,
                                                     build_version)
                            self.register_status(
                                bdoc,
                                "success",
                                job={"step": "release-note"},
                                publish={
                                    "full": {
                                        snapshot: {
                                            "conf": self.envconf,
                                            "release-note": {
                                                "base_dir": s3basedir,
                                                "bucket": s3_release_bucket,
                                                "url": urls
                                            }
                                        }
                                    }
                                })
                        except Exception as e:
                            self.logger.exception(
                                "Failed to upload release notes: %s" % e)
                            self.register_status(
                                bdoc,
                                "failed",
                                job={
                                    "step": "release-note",
                                    "err": str(e)
                                },
                                publish={
                                    "full": {
                                        snapshot: {
                                            "conf": self.envconf,
                                            "release-note": {
                                                "err": str(e),
                                                "base_dir": s3basedir,
                                                "bucket": s3_release_bucket
                                            },
                                        }
                                    }
                                })
                            raise

                    else:
                        self.logger.info(
                            "No release_folder found, no release notes will be part of the publishing"
                        )
                        # yet create the folder so we can dump metadata json file in there later
                        os.makedirs(release_folder)

                try:
                    self.register_status(bdoc,
                                         "publishing",
                                         transient=True,
                                         init=True,
                                         job={"step": "metadata"},
                                         publish={"full": {
                                             snapshot: {}
                                         }})
                    # now dump that metadata
                    build_info = "%s.json" % build_version
                    build_info_path = os.path.join(btconfig.RELEASE_PATH,
                                                   build_info)
                    json.dump(full_meta, open(build_info_path, "w"))
                    # override lastmodified header with our own timestamp
                    local_ts = dtparse(bdoc["_meta"]["build_date"])
                    utc_epoch = int(time.mktime(local_ts.timetuple()))
                    str_utc_epoch = str(utc_epoch)
                    # it's a full release, but all build info metadata (full, incremental) all go
                    # to the diff bucket (this is the main entry)
                    s3key = os.path.join(s3_release_folder, build_info)
                    aws.send_s3_file(
                        build_info_path,
                        s3key,
                        aws_key=self.envconf.get("cloud",
                                                 {}).get("access_key"),
                        aws_secret=self.envconf.get("cloud",
                                                    {}).get("secret_key"),
                        s3_bucket=s3_release_bucket,
                        metadata={"lastmodified": str_utc_epoch},
                        overwrite=True)
                    url = aws.get_s3_url(
                        s3key,
                        aws_key=self.envconf.get("cloud",
                                                 {}).get("access_key"),
                        aws_secret=self.envconf.get("cloud",
                                                    {}).get("secret_key"),
                        s3_bucket=s3_release_bucket)
                    self.logger.info(
                        "Full release metadata published for version: '%s'" %
                        url)
                    full_info = {
                        "build_version": full_meta["build_version"],
                        "require_version": None,
                        "target_version": full_meta["target_version"],
                        "type": full_meta["type"],
                        "release_date": full_meta["release_date"],
                        "url": url
                    }
                    publish_data_version(
                        s3_release_bucket,
                        s3_release_folder,
                        full_info,
                        aws_key=self.envconf.get("cloud",
                                                 {}).get("access_key"),
                        aws_secret=self.envconf.get("cloud",
                                                    {}).get("secret_key"))
                    self.logger.info("Registered version '%s'" %
                                     (build_version))
                    self.register_status(bdoc,
                                         "success",
                                         job={"step": "metadata"},
                                         publish={
                                             "full": {
                                                 snapshot: {
                                                     "conf": self.envconf,
                                                     "metadata": full_info
                                                 }
                                             }
                                         })
                except Exception as e:
                    self.logger.exception(
                        "Failed to upload snapshot metadata: %s" % e)
                    self.register_status(bdoc,
                                         "failed",
                                         job={
                                             "step": "metadata",
                                             "err": str(e)
                                         },
                                         publish={
                                             "full": {
                                                 snapshot: {
                                                     "conf": self.envconf,
                                                     "metadata": {
                                                         "err": str(e)
                                                     }
                                                 }
                                             }
                                         })
                    raise

            if "post" in steps:
                pinfo["step"] = "post"
                self.logger.info("Running post-publish step")
                self.register_status(bdoc,
                                     "post-publishing",
                                     transient=True,
                                     init=True,
                                     job={"step": "post-publish"},
                                     publish={"fulle": {
                                         snapshot: {}
                                     }})
                job = yield from self.job_manager.defer_to_thread(
                    pinfo,
                    partial(self.post_publish, snapshot, self.envconf, bdoc))
                job.add_done_callback(partial(done, step="post"))
                yield from job
                jobs.append(job)

            def published(f):
                try:
                    res = f.result()
                    self.logger.info("Snapshot '%s' uploaded to S3: %s" %
                                     (snapshot, res),
                                     extra={"notify": True})
                except Exception as e:
                    self.logger.exception(
                        "Failed to upload snapshot '%s' uploaded to S3: %s" %
                        (snapshot, e),
                        extra={"notify": True})

            if jobs:
                yield from asyncio.wait(jobs)
                task = asyncio.gather(*jobs)
                task.add_done_callback(published)
                yield from task


        def done(f):
            try:
                _ = f.result()
            except Exception as e:
                self.logger.exception("Unable to publish full release: %s" % e)
                raise

        task = asyncio.ensure_future(do())
        task.add_done_callback(done)

        return task

    def run_post_publish_snapshot(self, snapshot_name, repo_conf, build_doc):
        return self.run_pre_post("snapshot", "post", snapshot_name, repo_conf,
                                 build_doc)

    def post_publish(self, snapshot_name, repo_conf, build_doc):
        """
        Post-publish hook, running steps declared in config,
        but also whatever would be defined in a sub-class
        """
        return self.run_post_publish_snapshot(snapshot_name, repo_conf,
                                              build_doc)

    def pre_publish(self, snapshot_name, repo_conf, build_doc):
        """
        Pre-publish hook, running steps declared in config,
        but also whatever would be defined in a sub-class
        """
        return self.run_pre_publish_snapshot(snapshot_name, repo_conf,
                                             build_doc)


class DiffPublisher(BasePublisher):
    def __init__(self, diff_manager, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.diff_manager = diff_manager
        self.setup()

    def get_pre_post_previous_result(self, build_doc, key_value):
        previous_result = {
            "diff": build_doc["diff"][key_value]["diff"],
            "diff_folder": build_doc["diff"][key_value]["diff_folder"]
        }
        return previous_result

    def run_pre_publish_diff(self, previous_build_name, repo_conf, build_doc):
        return self.run_pre_post("diff", "pre", previous_build_name, repo_conf,
                                 build_doc)

    def pre_publish(self, previous_build_name, repo_conf, build_doc):
        """
        Pre-publish hook, running steps declared in config,
        but also whatever would be defined in a sub-class
        """
        return self.run_pre_publish_diff(previous_build_name, repo_conf,
                                         build_doc)

    def reset_synced(self, diff_folder, backend=None):
        """
        Remove "synced" flag from any pyobj file in diff_folder
        """
        synced_files = glob.glob(os.path.join(diff_folder, "*.pyobj.synced"))
        for synced in synced_files:
            diff_file = re.sub(r"\.pyobj\.synced$", ".pyobj", synced)
            os.rename(synced, diff_file)

    def get_release_note_filename(self, build_version):
        assert "." in build_version  # make sure it's an incremental
        _, tgt = build_version.split(".")
        return "release_%s" % tgt

    def publish(self,
                build_name,
                previous_build=None,
                steps=["pre", "reset", "upload", "meta", "post"]):
        """
        Publish diff files and metadata about the diff files, release note, etc... on s3.
        Using build_name, a src_build document is fetched, and a diff release is searched. If more
        than one diff release is found, "previous_build" must be specified to pick the correct one.
        - steps:
          * pre/post: optional steps processed as first and last steps.
          * reset: highly recommended, reset synced flag in diff files so they won't get skipped when used...
          * upload: upload diff_folder content to S3
          * meta: publish/register the version as available for auto-updating hubs
        """
        bdoc = self.load_build(build_name)
        assert bdoc, "No such build named '%s'" % build_name
        assert "diff" in bdoc, "No diff release found in build document named '%s'" % build_name
        if previous_build is None:
            dkeys = list(bdoc["diff"].keys())
            assert len(dkeys) == 1, "'previous_build' parameter is required because " \
                                    + "more than one diff release found: %s" % dkeys
            previous_build = dkeys.pop()
        assert previous_build, "No previous build could be found in order to pick correct diff release"
        # check what to do
        if type(steps) == str:
            steps = [steps]

        # instantiate publishing environment
        self.envconf = self.template_out_conf(bdoc)

        s3_release_folder = self.envconf["release"]["folder"]
        s3_release_bucket = self.envconf["release"]["bucket"]
        self.create_bucket(bucket_conf=self.envconf["release"],
                           credentials=self.envconf.get("cloud", {}))
        s3_diff_folder = self.envconf["diff"]["folder"]
        s3_diff_bucket = self.envconf["diff"]["bucket"]
        self.create_bucket(bucket_conf=self.envconf["diff"],
                           credentials=self.envconf.get("cloud", {}))

        # check whether a release note was generated for that (build_name,previous_build) couple
        release_folder = None
        if bdoc.get("release_note", {}).get(previous_build, {}):
            release_folder = bdoc["release_note"][previous_build][
                "release_folder"]

        diff_folder = bdoc["diff"][previous_build]["diff_folder"]
        try:
            meta = json.load(open(os.path.join(diff_folder, "metadata.json")))
        except FileNotFoundError:
            raise FileNotFoundError("metadata.json is missing")

        diff_version = meta["diff"]["version"]
        s3_diff_basedir = os.path.join(s3_diff_folder, diff_version)

        # hold error/exception on each step
        got_error = None

        @asyncio.coroutine
        def do():
            jobs = []
            pinfo = self.get_pinfo()
            pinfo["source"] = diff_folder
            pinfo["description"] = diff_version

            def done(f, step):
                try:
                    res = f.result()
                    self.register_status(bdoc,
                                         "success",
                                         job={
                                             "step": step,
                                             "result": res
                                         },
                                         publish={
                                             "incremental": {
                                                 previous_build: {
                                                     "conf": self.envconf,
                                                     step: res
                                                 }
                                             }
                                         })
                except Exception as e:
                    nonlocal got_error
                    got_error = e
                    self.register_status(bdoc,
                                         "failed",
                                         job={
                                             "step": step,
                                             "err": str(e)
                                         },
                                         publish={
                                             "incremental": {
                                                 previous_build: {
                                                     "conf": self.envconf,
                                                     step: {
                                                         "err": str(e)
                                                     }
                                                 }
                                             }
                                         })
                    self.logger.exception(
                        "Error while running %s-publish: %s" %
                        (step, got_error))

            if "_meta" not in bdoc:
                raise PublisherException("No metadata (_meta) found in build document")

            if "pre" in steps:
                # then we upload all the folder content
                pinfo["step"] = "pre"
                self.logger.info("Running pre-publish step")
                self.register_status(
                    bdoc,
                    "pre",
                    transient=True,
                    init=True,
                    job={"step": "pre"},
                    publish={"incremental": {
                        previous_build: {}
                    }})
                job = yield from self.job_manager.defer_to_thread(
                    pinfo,
                    partial(self.pre_publish, previous_build, self.envconf,
                            bdoc))
                job.add_done_callback(partial(done, step="pre"))
                yield from job
                if got_error:
                    raise got_error
                jobs.append(job)

            if "reset" in steps:
                # first we need to reset "synced" flag in diff files to make
                # sure all of them will be applied by client
                pinfo["step"] = "reset synced"
                self.logger.info(
                    "Resetting 'synced' flag in pyobj files located in folder '%s'"
                    % diff_folder)
                self.register_status(
                    bdoc,
                    "reset-synced",
                    transient=True,
                    init=True,
                    job={"step": "reset-synced"},
                    publish={"incremental": {
                        previous_build: {}
                    }})
                job = yield from self.job_manager.defer_to_thread(
                    pinfo, partial(self.reset_synced, diff_folder))
                job.add_done_callback(partial(done, step="reset"))
                yield from job
                if got_error:
                    raise got_error
                jobs.append(job)

            if "upload" in steps:
                # then we upload all the folder content
                pinfo["step"] = "upload"
                self.logger.info(
                    "Uploading files from '%s' to s3 (%s/%s)" %
                    (diff_folder, s3_diff_bucket, s3_diff_basedir))
                self.register_status(
                    bdoc,
                    "upload",
                    transient=True,
                    init=True,
                    job={"step": "upload"},
                    publish={"incremental": {
                        previous_build: {}
                    }})
                job = yield from self.job_manager.defer_to_thread(
                    pinfo,
                    partial(aws.send_s3_folder,
                            diff_folder,
                            s3basedir=s3_diff_basedir,
                            aws_key=self.envconf.get("cloud",
                                                     {}).get("access_key"),
                            aws_secret=self.envconf.get("cloud",
                                                        {}).get("secret_key"),
                            s3_bucket=s3_diff_bucket,
                            overwrite=True))
                job.add_done_callback(partial(done, step="upload"))
                yield from job
                jobs.append(job)

            if "meta" in steps:
                # finally we create a metadata json file pointing to this release
                try:
                    self.register_status(
                        bdoc,
                        "metadata",
                        transient=True,
                        init=True,
                        job={"step": "metadata"},
                        publish={"incremental": {
                            previous_build: {}
                        }})
                    pinfo["step"] = "generate meta"
                    self.logger.info(
                        "Generating JSON metadata for incremental release '%s'"
                        % diff_version)
                    # if the same, this would create an infinite loop in autoupdate hub
                    # (X requires X, where to find X ? there, but X requires X, where to find X ?...)
                    if meta["old"]["version"] == meta["new"]["version"]:
                        raise PublisherException("Required version is the same as target version "
                                                 + "('%s'), prevent publishing to avoid infinite loop " % meta["new"]["version"]
                                                 + "while resolving updates in auto-update hub")
                    # generate json metadata about this diff release
                    if getattr(btconfig, "SKIP_CHECK_VERSIONS", None):
                        self.logger.info(
                            "SKIP_CHECK_VERSIONS %s, no version check will be performed on diff metadata"
                            % repr(btconfig.SKIP_CHECK_VERSIONS))
                    else:
                        assert getattr(
                            btconfig, "BIOTHINGS_VERSION", "master"
                        ) != "master", "I won't publish data refering BIOTHINGS_VERSION='master'"
                        assert getattr(
                            btconfig, "APP_VERSION", "master"
                        ) != "master", "I won't publish data refering APP_VERSION='master'"
                        assert getattr(btconfig, "STANDALONE_VERSION",
                                       None), "STANDALONE_VERSION not defined"
                    diff_meta = {
                        "type": "incremental",
                        "build_version": diff_version,
                        "require_version": meta["old"]["version"],
                        "target_version": meta["new"]["version"],
                        "release_date": datetime.now().isoformat(),
                        "app_version": btconfig.APP_VERSION,
                        "biothings_version": btconfig.BIOTHINGS_VERSION,
                        "standalone_version": btconfig.STANDALONE_VERSION,
                        "metadata": {
                            "url":
                            aws.get_s3_url(
                                os.path.join(s3_diff_basedir, "metadata.json"),
                                aws_key=self.envconf.get("cloud",
                                                         {}).get("access_key"),
                                aws_secret=self.envconf.get(
                                    "cloud", {}).get("secret_key"),
                                s3_bucket=s3_diff_bucket)
                        },
                    }
                    # upload release notes
                    if release_folder:
                        urls = self.publish_release_notes(
                            release_folder,
                            diff_version,
                            s3_release_folder,
                            s3_release_bucket,
                            aws_key=self.envconf.get("cloud",
                                                     {}).get("access_key"),
                            aws_secret=self.envconf.get("cloud",
                                                        {}).get("secret_key"))
                        diff_meta.setdefault("changes", {})
                        diff_meta["changes"].update(urls)
                    # and then upload diff metadataf files
                    diff_file = "%s.json" % diff_version
                    diff_meta_path = os.path.join(btconfig.RELEASE_PATH,
                                                  diff_file)
                    json.dump(diff_meta,
                              open(diff_meta_path, "w"),
                              indent=True)
                    # get a timestamp from metadata to force lastdmodifed header
                    # timestamp is when the new collection was built (not when the diff
                    # was generated, as diff can be generated way after). New collection's
                    # timestamp remains a good choice as data (diff) relates to that date anyway
                    local_ts = dtparse(diff_meta["release_date"])
                    utc_epoch = int(time.mktime(local_ts.timetuple()))
                    str_utc_epoch = str(utc_epoch)
                    s3key = os.path.join(s3_release_folder, diff_file)
                    aws.send_s3_file(
                        diff_meta_path,
                        s3key,
                        aws_key=self.envconf.get("cloud",
                                                 {}).get("access_key"),
                        aws_secret=self.envconf.get("cloud",
                                                    {}).get("secret_key"),
                        s3_bucket=s3_release_bucket,
                        metadata={"lastmodified": str_utc_epoch},
                        overwrite=True)
                    url = aws.get_s3_url(
                        s3key,
                        aws_key=self.envconf.get("cloud",
                                                 {}).get("access_key"),
                        aws_secret=self.envconf.get("cloud",
                                                    {}).get("secret_key"),
                        s3_bucket=s3_release_bucket)
                    self.logger.info(
                        "Incremental release metadata published for version: '%s'"
                        % url)
                    diff_info = {
                        "build_version": diff_meta["build_version"],
                        "require_version": diff_meta["require_version"],
                        "target_version": diff_meta["target_version"],
                        "type": diff_meta["type"],
                        "release_date": diff_meta["release_date"],
                        "url": url
                    }
                    publish_data_version(
                        s3_release_bucket,
                        s3_release_folder,
                        diff_info,
                        aws_key=self.envconf.get("cloud",
                                                 {}).get("access_key"),
                        aws_secret=self.envconf.get("cloud",
                                                    {}).get("secret_key"))
                    self.logger.info("Registered version '%s'" %
                                     (diff_version))

                    self.register_status(bdoc,
                                         "success",
                                         job={"step": "metadata"},
                                         publish={
                                             "incremental": {
                                                 previous_build: {
                                                     "conf": self.envconf,
                                                     "metadata": diff_info
                                                 }
                                             }
                                         })

                except Exception as e:
                    self.logger.exception(
                        "Failed to upload snapshot metadata: %s" % e)
                    self.register_status(bdoc,
                                         "failed",
                                         job={
                                             "step": "metadata",
                                             "err": str(e)
                                         },
                                         publish={
                                             "incremental": {
                                                 previous_build: {
                                                     "conf": self.envconf,
                                                     "metadata": {
                                                         "err": str(e)
                                                     }
                                                 }
                                             }
                                         })
                    raise

                jobs.append(job)

            if "post" in steps:
                pinfo["step"] = "post"
                self.logger.info("Running post-publish step")
                self.register_status(
                    bdoc,
                    "post",
                    transient=True,
                    init=True,
                    job={"step": "post"},
                    publish={"incremental": {
                        previous_build: {}
                    }})
                job = yield from self.job_manager.defer_to_thread(
                    pinfo,
                    partial(self.post_publish, previous_build, self.envconf,
                            bdoc))
                job.add_done_callback(partial(done, step="post"))
                yield from job
                if got_error:
                    raise got_error
                jobs.append(job)

            def uploaded(f):
                try:
                    res = f.result()
                    self.logger.info("Diff folder '%s' uploaded to S3: %s" %
                                     (diff_folder, res),
                                     extra={"notify": True})
                except Exception as e:
                    self.logger.exception(
                        "Failed to upload diff folder '%s' uploaded to S3: %s"
                        % (diff_folder, e),
                        extra={"notify": True})

            yield from asyncio.wait(jobs)
            task = asyncio.gather(*jobs)
            task.add_done_callback(uploaded)
            yield from task

        def done(f):
            try:
                _ = f.result()
            except Exception as e:
                self.logger.exception("Unable to publish incremental release: %s" % e)
                raise

        task = asyncio.ensure_future(do())
        task.add_done_callback(done)

        return task

    def run_post_publish_diff(self, build_name, repo_conf, build_doc):
        return self.run_pre_post("diff", "post", build_name, repo_conf,
                                 build_doc)

    def post_publish(self, build_name, repo_conf, build_doc):
        """
        Post-publish hook, running steps declared in config,
        but also whatever would be defined in a sub-class
        """
        return self.run_post_publish_diff(build_name, repo_conf, build_doc)


class ReleaseManager(BaseManager, BaseStatusRegisterer):

    DEFAULT_SNAPSHOT_PUBLISHER_CLASS = SnapshotPublisher
    DEFAULT_DIFF_PUBLISHER_CLASS = DiffPublisher

    def __init__(self,
                 diff_manager,
                 snapshot_manager,
                 poll_schedule=None,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.diff_manager = diff_manager
        self.snapshot_manager = snapshot_manager
        self.t0 = time.time()
        self.log_folder = btconfig.LOG_FOLDER
        self.timestamp = datetime.now()
        self.release_config = {}
        self.poll_schedule = poll_schedule
        self.es_backups_folder = getattr(btconfig, "ES_BACKUPS_FOLDER", None)
        self.setup()

    def clean_stale_status(self):
        src_build = get_src_build()
        for build in src_build.find():
            dirty = False
            for job in build.get("jobs", []):
                if job.get("status") == "generating":
                    logging.warning(
                        "Found stale build '%s', marking release-note status as 'canceled'"
                        % build["_id"])
                    job["status"] = "canceled"
                    dirty = True
            if dirty:
                src_build.replace_one({"_id": build["_id"]}, build)

    def setup(self):
        self.setup_log()

    def setup_log(self):
        self.logger, self.logfile = get_logger(RELEASEMANAGER_CATEGORY,
                                               self.log_folder)

    def poll(self, state, func):
        super().poll(state, func, col=get_src_build())

    def __getitem__(self, stage_env):
        """
        Return an instance of a releaser for the release environment named "env"
        """
        stage, env = stage_env
        kwargs = copy.copy(self.register[stage_env])
        klass = kwargs.pop("class")
        return klass(**kwargs)

    def configure(self, release_confdict):
        """
        Configure manager with release "confdict". See config_hub.py in API
        for the format.
        """
        self.release_config = copy.deepcopy(release_confdict)
        for env, envconf in self.release_config.get("env", {}).items():
            try:
                if envconf.get("cloud"):
                    assert envconf["cloud"]["type"] == "aws", \
                        "Only Amazon AWS cloud is supported at the moment"
                # here we register publisher class and args passed during init
                # which are common to the manager
                self.register[("snapshot", env)] = {
                    "class": self.DEFAULT_SNAPSHOT_PUBLISHER_CLASS,
                    "envconf": envconf,
                    "job_manager": self.job_manager,
                    "snapshot_manager": self.snapshot_manager,
                    "log_folder": self.log_folder,
                    "es_backups_folder": self.es_backups_folder,
                }
                self.register[("diff", env)] = {
                    "class": self.DEFAULT_DIFF_PUBLISHER_CLASS,
                    "envconf": envconf,
                    "job_manager": self.job_manager,
                    "diff_manager": self.diff_manager,
                    "log_folder": self.log_folder,
                    "es_backups_folder": self.es_backups_folder,
                }
            except Exception as e:
                self.logger.exception(
                    "Couldn't setup release environment '%s' because: %s" %
                    (env, e))

    def get_predicates(self):
        return []

    def get_pinfo(self):
        """
        Return dict containing information about the current process
        (used to report in the hub)
        """
        pinfo = {
            "category": RELEASEMANAGER_CATEGORY,
            "source": "",
            "step": "",
            "description": ""
        }
        preds = self.get_predicates()
        if preds:
            pinfo["__predicates__"] = preds
        return pinfo

    @property
    def collection(self):
        return get_src_build()

    def register_status(self,
                        bdoc,
                        stage,
                        status,
                        transient=False,
                        init=False,
                        **extra):
        BaseStatusRegisterer.register_status(self,
                                             bdoc,
                                             stage,
                                             status,
                                             transient=transient,
                                             init=init,
                                             **extra)

    def reset_synced(self, old, new):
        """
        Reset sync flags for diff files produced between "old" and "new" build.
        Once a diff has been applied, diff files are flagged as synced so subsequent diff
        won't be applied twice (for optimization reasons, not to avoid data corruption since
        diff files can be safely applied multiple times).
        In any needs to apply the diff another time, diff files needs to reset.
        """
        # we need a diff publisher to do that, whatever the target (s3, filesystem,...)
        dtypes = [pub for pub in self.register if pub[0] == "diff"]
        assert len(
            dtypes
        ), "Can't reset synced diff files, no diff publisher registered"
        # get one, all are able to reset synced files
        dtype = dtypes[0]
        diff_folder = generate_folder(btconfig.DIFF_PATH, old, new)
        diff_publisher = self[dtype]
        return diff_publisher.reset_synced(diff_folder)

    def load_build(self, key_name, stage=None):
        if stage is None:
            # picke build doc searching for snapshot then diff key if not found
            return self.load_doc(key_name, "snapshot") or self.load_doc(
                key_name, "diff")
        else:
            return self.load_doc(key_name, stage)

    def publish_diff(self,
                     publisher_env,
                     build_name,
                     previous_build=None,
                     steps=["pre", "reset", "upload", "meta", "post"]):
        if publisher_env not in self.release_config.get("env", {}):
            raise ValueError("Unknonw release environment '%s'" %
                             publisher_env)
        publisher = self[("diff", publisher_env)]
        return publisher.publish(build_name=build_name,
                                 previous_build=previous_build,
                                 steps=steps)

    def publish_snapshot(self,
                         publisher_env,
                         snapshot,
                         build_name=None,
                         previous_build=None,
                         steps=["pre", "meta", "post"]):
        if publisher_env not in self.release_config.get("env", {}):
            raise ValueError("Unknonw release environment '%s'" %
                             publisher_env)
        publisher = self[("snapshot", publisher_env)]
        return publisher.publish(snapshot=snapshot,
                                 build_name=build_name,
                                 previous_build=previous_build,
                                 steps=steps)

    def publish(self, publisher_env, snapshot_or_build_name, *args, **kwargs):
        snapshot_doc = None
        diff_doc = None
        try:
            snapshot_doc = self.load_doc(snapshot_or_build_name, "snapshot")
            if isinstance(snapshot_doc, list):
                raise PublisherException("More than one build_doc associated to snapshot '%s', " % snapshot_or_build_name + "use publish_snapshot()")
        except AssertionError:
            # no doc at all
            pass
        try:
            diff_doc = self.load_doc(snapshot_or_build_name, "diff")
            if isinstance(diff_doc, list):
                raise PublisherException("More than one build_doc associated to diff '%s', " % snapshot_or_build_name + "use explicetely publish_diff()")
        except AssertionError:
            # no doc at all
            pass
        # check returned doc contains what we think it does and load_doc hasn't use snapshot_or_build_name as _id
        if snapshot_doc and snapshot_or_build_name not in snapshot_doc.get(
                "snapshot", {}):
            # doc was returend with snapshot_or_build_name matching _id, not snapshot name, invalidate
            snapshot_doc = None
        if diff_doc and snapshot_or_build_name not in diff_doc.get("diff", {}):
            diff_doc = None
        # do we still have something ambiguous ?
        if snapshot_doc and diff_doc:
            # so we have 2 releases associated, we can't know which one user wants
            raise PublisherException("'%s' is associated to 2 different releases " % snapshot_or_build_name
                                     + "(document _id '%s' and '%s'" % (snapshot_doc["_id"], diff_doc["_id"])
                                     + "use publish_snapshot() or publish_diff()")
        elif snapshot_doc:
            self.logger.info("'%s' associated to a snapshot/full release" %
                             snapshot_or_build_name)
            return self.publish_snapshot(publisher_env, snapshot_or_build_name,
                                         *args, **kwargs)
        elif diff_doc:
            self.logger.info("'%s' associated to a diff/incremental release" %
                             snapshot_or_build_name)
            return self.publish_diff(publisher_env, snapshot_or_build_name,
                                     *args, **kwargs)
        else:
            raise PublisherException("No release associated to '%s'" %
                                     snapshot_or_build_name)

    def get_release_note(self, old, new, format="txt", prefix="release_*"):
        release_folder = generate_folder(btconfig.RELEASE_PATH, old, new)
        if not os.path.exists(release_folder):
            raise PublisherException("No release note folder found")
        notes = glob.glob(
            os.path.join(release_folder, "%s.%s" % (prefix, format)))
        if not notes:
            raise PublisherException("No release notes found in folder")
        if len(notes) != 1:
            raise PublisherException(
                "Found %d notes (%s), expected only one" %
                (len(notes), [os.path.basename(n) for n in notes]))
        content = open(notes[0]).read()
        if format == "json":
            content = json.loads(content)
        return content

    def create_release_note(self,
                            old,
                            new,
                            filename=None,
                            note=None,
                            format="txt"):
        """
        Generate release note files, in TXT and JSON format, containing significant changes
        summary between target collections old and new. Output files
        are stored in a diff folder using generate_folder(old,new).

        'filename' can optionally be specified, though it's not recommended as the publishing pipeline,
        using these files, expects a filenaming convention.

        'note' is an optional free text that can be added to the release note, at the end.

        txt 'format' is the only one supported for now.
        """
        bdoc = self.load_build(new)
        old = old or get_previous_collection(new)
        release_folder = generate_folder(btconfig.RELEASE_PATH, old, new)
        if not os.path.exists(release_folder):
            os.makedirs(release_folder)
        filepath = None

        def do():
            changes = self.build_release_note(old, new, note=note)
            nonlocal filepath
            nonlocal filename
            assert format == "txt", "Only 'txt' format supported for now"
            filename = filename or "release_%s.%s" % (
                changes["new"]["_version"], format)
            filepath = os.path.join(release_folder, filename)
            render = ReleaseNoteTxt(changes)
            txt = render.save(filepath)
            filename = filename.replace(".%s" % format, ".json")
            filepath = os.path.join(release_folder, filename)
            json.dump(changes, open(filepath, "w"), indent=True)
            return {"txt": txt, "changes": changes}

        @asyncio.coroutine
        def main(release_folder):
            got_error = False
            pinfo = self.get_pinfo()
            pinfo["step"] = "release_note"
            pinfo["source"] = release_folder
            pinfo["description"] = filename
            self.register_status(bdoc,
                                 "release_note",
                                 "generating",
                                 transient=True,
                                 init=True,
                                 job={"step": "release_note"},
                                 release_note={old: {}})
            job = yield from self.job_manager.defer_to_thread(pinfo, do)

            def reported(f):
                nonlocal got_error
                try:
                    res = f.result()
                    self.register_status(bdoc,
                                         "release_note",
                                         "success",
                                         job={"step": "release_note"},
                                         release_note={
                                             old: {
                                                 "changes": res["changes"],
                                                 "release_folder":
                                                 release_folder
                                             }
                                         })
                    self.logger.info("Release note ready, saved in %s: %s" %
                                     (release_folder, res["txt"]),
                                     extra={"notify": True})
                    set_pending_to_publish(new)
                except Exception as e:
                    self.logger.exception(e)
                    got_error = e

            job.add_done_callback(reported)
            yield from job
            if got_error:
                self.logger.exception("Failed to create release note: %s" %
                                      got_error,
                                      extra={"notify": True})
                self.register_status(bdoc,
                                     "release_note",
                                     "failed",
                                     job={
                                         "step": "release_note",
                                         "err": str(e)
                                     },
                                     release_note={old: {}})
                raise got_error

        job = asyncio.ensure_future(main(release_folder))
        return job

    def build_release_note(self, old_colname, new_colname, note=None):
        """
        Build a release note containing most significant changes between build names "old_colname" and "new_colname".
        An optional end note can be added to bring more specific information about the release.

        Return a dictionnary containing significant changes.
        """
        def get_counts(dstats):
            stats = {}
            for subsrc, count in dstats.items():
                try:
                    src_sub = get_source_fullname(subsrc).split(".")
                except AttributeError:
                    # not a merge stats coming from a source
                    # (could be custom field stats, eg. total_* in mygene)
                    src_sub = [subsrc]
                if len(src_sub) > 1:
                    # we have sub-sources we need to split the count
                    src, sub = src_sub
                    stats.setdefault(src, {})
                    stats[src][sub] = {"_count": count}
                else:
                    src = src_sub[0]
                    stats[src] = {"_count": count}
            return stats

        def get_versions(doc):
            try:
                versions = dict((k, {"_version": v["version"]}) for k, v in
                                doc.get("_meta", {}).get("src", {}).items() if "version" in v)
            except KeyError:
                # previous version format
                versions = dict((k, {
                    "_version": v
                })
                    for k, v in doc.get("_meta", {}).get(
                        "src_version", {}).items())
            return versions

        diff_folder = generate_folder(btconfig.DIFF_PATH, old_colname,
                                      new_colname)
        try:
            metafile = os.path.join(diff_folder, "metadata.json")
            metadata = json.load(open(metafile))
            old_colname = metadata["old"]["backend"]
            new_colname = metadata["new"]["backend"]
            diff_stats = metadata["diff"]["stats"]
        except FileNotFoundError:
            # we're generating a release note without diff information
            self.logger.info(
                "No metadata.json file found, this release note won't have diff stats included"
            )
            diff_stats = {}

        new = create_backend(new_colname)
        old = create_backend(old_colname)
        assert isinstance(old, DocMongoBackend) and isinstance(new, DocMongoBackend), \
            "Only MongoDB backend types are allowed when generating a release note"
        assert old.target_collection.database.name == btconfig.DATA_TARGET_DATABASE and \
            new.target_collection.database.name == btconfig.DATA_TARGET_DATABASE, \
            "Target databases must match current DATA_TARGET_DATABASE setting"
        new_doc = get_src_build().find_one({"_id": new.target_collection.name})
        if not new_doc:
            raise PublisherException("Collection '%s' has no corresponding build document" %
                                     new.target_collection.name)
        # old_doc doesn't have to exist (but new_doc has) in case we build a initial release note
        # compared against nothing
        old_doc = get_src_build().find_one({"_id": old.target_collection.name
                                            }) or {}
        tgt_db = get_target_db()
        old_total = tgt_db[old.target_collection.name].count()
        new_total = tgt_db[new.target_collection.name].count()
        changes = {
            "old": {
                "_version": old.version,
                "_count": old_total,
            },
            "new": {
                "_version": new.version,
                "_count": new_total,
                "_fields": {},
                "_summary": diff_stats,
            },
            "stats": {
                "added": {},
                "deleted": {},
                "updated": {},
            },
            "note": note,
            "generated_on": str(datetime.now()),
            "sources": {
                "added": {},
                "deleted": {},
                "updated": {},
            }
        }
        # for later use
        new_versions = get_versions(new_doc)
        old_versions = get_versions(old_doc)
        # now deal with stats/counts. Counts are related to uploader, ie. sub-sources
        new_merge_stats = get_counts(new_doc.get("merge_stats", {}))
        old_merge_stats = get_counts(old_doc.get("merge_stats", {}))
        new_stats = get_counts(new_doc.get("_meta", {}).get("stats", {}))
        old_stats = get_counts(old_doc.get("_meta", {}).get("stats", {}))
        new_info = update_dict_recur(new_versions, new_merge_stats)
        old_info = update_dict_recur(old_versions, old_merge_stats)

        def analyze_diff(ops, destdict, old, new):
            for op in ops:
                # get main source / main field
                key = op["path"].strip("/").split("/")[0]
                if op["op"] == "add":
                    destdict["added"][key] = new[key]
                elif op["op"] == "remove":
                    destdict["deleted"][key] = old[key]
                elif op["op"] == "replace":
                    destdict["updated"][key] = {
                        "new": new[key],
                        "old": old[key]
                    }
                else:
                    raise ValueError(
                        "Unknown operation '%s' while computing changes" %
                        op["op"])

        # diff source info
        # this only works on main source information: if there's a difference in a
        # sub-source, it won't be shown but considered as if it was the main-source
        ops = jsondiff(old_info, new_info)
        analyze_diff(ops, changes["sources"], old_info, new_info)

        ops = jsondiff(old_stats, new_stats)
        analyze_diff(ops, changes["stats"], old_stats, new_stats)

        # mapping diff: we re-compute them and don't use any mapping.pyobj because that file
        # only allows "add" operation as a safety rule (can't delete fields in ES mapping once indexed)
        ops = jsondiff(old_doc.get("mapping", {}), new_doc["mapping"])
        for op in ops:
            changes["new"]["_fields"].setdefault(op["op"], []).append(
                op["path"].strip("/").replace("/", "."))

        return changes

    def release_info(self, env=None, remote=False):
        res = copy.deepcopy(self.release_config)
        for kenv in self.release_config["env"]:
            if env and env != kenv:
                continue
            if remote:
                raise NotImplementedError()
                # TODO: could list all releases from S3 ?
        return res



from tornado.web import HTTPError

class BadRequest(HTTPError):

    def __init__(self, log_message=None, *args, **kwargs):
        super().__init__(400, log_message, *args, **kwargs)
        self.kwargs = dict(kwargs) or {}
        self.kwargs.pop('reason', None)

class EndRequest(HTTPError):
    """
        Similar to tornado.web.Finish() but write_error handles it.
    """

    def __init__(self, status_code=200, log_message=None, *args, **kwargs):
        super().__init__(status_code, log_message, *args, **kwargs)
        self.kwargs = dict(kwargs) or {}
        self.kwargs.pop('reason', None)




import importlib.util
import inspect
import logging
import os
import types
from importlib import import_module
from pprint import pformat
from pydoc import locate

import tornado.log
from tornado.web import Application

from biothings.web.settings import default as web_default
from biothings.web.handlers import BaseAPIHandler, BaseESRequestHandler
from biothings.web.options import OptionSets

try:
    from raven.contrib.tornado import AsyncSentryClient
except ImportError:
    __SENTRY_INSTALLED__ = False
else:
    __SENTRY_INSTALLED__ = True

class BiothingConfigError(Exception):
    pass

class BiothingWebSettings():
    '''
    A container for the settings that configure the web API.

    * Environment variables can override settings of the same names.
    * Default values are defined in biothings.web.settings.default.

    '''
    _default = object()
    _parent = object()
    _user = object()

    def __init__(self, config=None, parent=None, **kwargs):
        '''
        :param config: a module that configures this biothing
            or its fully qualified name,
            or its module file path.
        '''
        self._default = web_default
        self._parent = parent  # config package
        self._user = self.load_module(config, object())
        self.logger.info("%s", self._user)  # log file location

        # process keyword setting override
        for key, value in kwargs.items():
            setattr(self._user, key, value)

        # process environment variable override of named settings
        for name in os.environ:
            if hasattr(self, name):
                new_value = None
                if isinstance(getattr(self, name), str):
                    new_value = os.environ[name]
                elif isinstance(getattr(self, name), bool):
                    new_value = os.environ[name].lower() in ('true', '1')
                if new_value is not None:
                    self.logger.info("$ %s = %s", name, os.environ[name])
                    setattr(self._user, name, new_value)
                else:  # cannot override dict, array, object type...
                    self.logger.error("Env %s is not suppored.", name)

        # for metadata dev details
        if os.path.isdir(os.path.join(self.APP_GIT_REPOSITORY, '.git')):
            self._git_repo_path = self.APP_GIT_REPOSITORY
        else:
            self._git_repo_path = None

        self.optionsets = OptionSets()
        self.handlers = {}
        self.validate()

    @staticmethod
    def load_module(config, default=None):
        """
        Ensure config is a module.
        If config does not evaluate,
        Return default if it's provided.
        """
        if isinstance(config, types.ModuleType):
            return config
        elif isinstance(config, str) and config.endswith('.py'):
            spec = importlib.util.spec_from_file_location("config", config)
            config = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(config)
            return config
        elif isinstance(config, str) and config:
            return import_module(config)
        elif not config and default:
            return default
        raise BiothingConfigError()

    def __getattr__(self, name):
        if hasattr(self._user, name):
            return getattr(self._user, name)
        elif hasattr(self._parent, name):
            return getattr(self._parent, name)
        elif hasattr(self._default, name):
            return getattr(self._default, name)
        else:  # not provided and no default
            raise AttributeError()

    @property
    def logger(self):
        return logging.getLogger('biothings.web.settings')

    @staticmethod
    def load_class(kls):
        """
        Ensure config is a module.
        If config does not evaluate,
        Return default if it's provided.
        """
        if inspect.isclass(kls):
            return kls
        if isinstance(kls, str):
            return locate(kls)
        raise BiothingConfigError()

    def _generate_app_settings(self, override=None):
        """
        Generates settings for tornado.web.Application. This result and the
        method below can define a tornado application to start a web server.
        """
        settings = {
            'biothings': self,
            'autoreload': False,
            'debug': False,
        }
        settings.update(override or {})
        supported_keywords = (
            'default_handler_class', 'default_handler_args', 'template_path',
            'log_function', 'compress_response', 'cookie_secret',
            'login_url', 'static_path', 'static_url_prefix')

        for setting in supported_keywords:
            if hasattr(self, setting.upper()):
                settings[setting] = getattr(self, setting.upper())

        if __SENTRY_INSTALLED__ and self.SENTRY_CLIENT_KEY:
            # Setup error monitoring with Sentry. More on:
            # https://docs.sentry.io/clients/python/integrations/#tornado
            settings['sentry_client'] = AsyncSentryClient(self.SENTRY_CLIENT_KEY)

        return settings

    def _generate_app_handlers(self, addons=None):
        '''
        Generates the tornado.web.Application `(regex, handler_class, options) tuples
        <http://www.tornadoweb.org/en/stable/web.html#application-configuration>`_.
        '''
        handlers = {}
        addons = addons or []
        for rule in self.APP_LIST + addons:
            pattern = rule[0]
            handler = self.load_class(rule[1])
            setting = rule[2] if len(rule) == 3 else {}
            assert handler, rule[1]
            if issubclass(handler, BaseAPIHandler) and handler.name:
                handler_name = handler.name
                handler_options = handler.kwargs
                setting_attr = '_'.join((handler_name, 'kwargs')).upper()
                setting_options = getattr(self, setting_attr, {})
                self.optionsets.add(handler_name, handler_options)
                self.optionsets.add(handler_name, setting_options)
                self.optionsets.groups[handler_name] = handler.kwarg_groups
                self.optionsets.methods[handler_name] = handler.kwarg_methods
            if '{typ}' in pattern:
                if not issubclass(handler, BaseESRequestHandler):
                    raise BiothingConfigError()
                for biothing_type in self.BIOTHING_TYPES:
                    _pattern = pattern.format(
                        pre=self.API_PREFIX,
                        ver=self.API_VERSION,
                        typ=biothing_type).replace('//', '/')
                    _setting = dict(setting)
                    _setting['biothing_type'] = biothing_type
                    handlers[_pattern] = (_pattern, handler, _setting)
            elif '{pre}' in pattern or '{ver}' in pattern:
                pattern = pattern.format(
                    pre=self.API_PREFIX,
                    ver=self.API_VERSION).replace('//', '/')
                handlers[pattern] = (pattern, handler, setting)
            else:  # no pattern translation
                handlers[pattern] = (pattern, handler, setting)

        self.handlers = handlers
        handlers = list(handlers.values())
        self.logger.info('API Handlers:\n%s', pformat(handlers, width=200))
        return handlers

    def get_app(self, settings=False, handlers=None):
        """
        Return the tornado.web.Application defined by this settings.
        This is primarily how an HTTP server interacts with this class.
        Additional settings and handlers accepted as parameters.
        """
        # config package
        if self._user.__name__ == self._user.__package__:
            attrs = [getattr(self._user, attr) for attr in dir(self._user)]
            confs = [attr for attr in attrs if isinstance(attr, types.ModuleType)]
            _settings = [self.__class__(_attr, self._user) for _attr in confs]
            _handlers = [(f'/{c.API_PREFIX}/.*', c.get_app(settings)) for c in _settings]
            _handlers += handlers or []  # second level front pages won't be exposed
        else:  # config module
            _handlers = self._generate_app_handlers(handlers)
        _settings = self._generate_app_settings(settings)
        return Application(_handlers, **_settings)

    def get_git_repo_path(self):
        '''
        Return the path of the codebase if the specified folder in settings exists or `None`.
        '''
        return self._git_repo_path

    def configure_logger(self, logger):
        '''
        Configure a logger's formatter to use the format defined in this web setting.
        '''
        try:
            if self.LOGGING_FORMAT and logger.hasHandlers():
                for handler in logger.handlers:
                    if isinstance(handler.formatter, tornado.log.LogFormatter):
                        handler.formatter._fmt = self.LOGGING_FORMAT
        except Exception:
            self.logger.exception('Error configuring logger %s.', logger)

    def validate(self):
        '''
        Validate the settings defined for this web server.
        '''
        assert self.API_VERSION or self.API_PREFIX
        assert isinstance(self.LIST_SIZE_CAP, int)
        assert isinstance(self.ACCESS_CONTROL_ALLOW_METHODS, str)
        assert isinstance(self.ACCESS_CONTROL_ALLOW_HEADERS, str)



"""
    Biothings Test Case Helper

    Envs:
    TEST_TIMEOUT    Individual request timeout in seconds.
    TEST_HOST       Tornado API server URL to test on. For example:

                    - When not specified, starts a local server
                    - Test a remote API server: http://www.mygene.info/v3
                    - Test a local API server: http://localhost:8000/api

    ES_HOST         Elasticsearch host address. Default to localhost:9200.

"""
import os
from functools import partial

import pytest
import requests
from tornado.ioloop import IOLoop
from tornado.testing import AsyncHTTPTestCase

from biothings.web.settings import BiothingESWebSettings


class BiothingsTestCase(AsyncHTTPTestCase):
    """
        Starts a tornado server to run tests locally.
        Need a config.py under the current working dir.
        If a host is specified, test against that host.
    """
    conf = 'config'
    host = os.getenv("TEST_HOST", '')  # test locally when empty.
    path = ''  # api path with prefix, populated from web settings.

    @classmethod
    def setup_class(cls):
        cls.settings = BiothingESWebSettings(cls.conf)
        prefix = cls.settings.API_PREFIX
        version = cls.settings.API_VERSION
        cls.path = f'/{prefix}/{version}/'
        cls.path = cls.path.replace('//', '/')
        cls.host = cls.host.rstrip('/')

    # override
    def get_new_ioloop(self):
        return IOLoop.current()

    # override
    def get_app(self):
        return self.settings.get_app()

    # override
    def request(self, path, method="GET", expect=200, *args, **kwargs):
        ''' Use requests.py instead of the built-in client
            Override to make the requests non-blocking
            param: path: network path to make request to
            param: method: ('GET') http request method
            param: expect: (200) check status code
        '''
        url = self.get_url(path)

        if self.host:  # remote test
            res = requests.request(method, url, **kwargs)

        else:  # local test
            func = partial(requests.request, method, url, **kwargs)
            res = self.io_loop.run_sync(
                lambda: self.io_loop.run_in_executor(None, func, *args),
                timeout=os.getenv("TEST_TIMEOUT"))

        assert res.status_code == expect, res.text
        return res

    # override
    def get_url(self, path):
        """
        Return the URL that can be passed to an HTTP client.

        http://example.com/     ->      http://example.com/
        /status                 ->      http://<test_server>/status
        query?q=cdk2            ->      http://<test_server>/<api_path>/query?q=cdk2
        """
        if path.lower().startswith(("http://", "https://")):
            return path
        if not path.startswith('/'):  # biothings api call
            path = self.path + path
        if self.host:                 # remote server
            return self.host + path
        return super().get_url(path)  # local server

    def query(self, method='GET', endpoint='query', hits=True, data=None, json=None, **kwargs):
        """ Make a query and assert positive hits by default.
            Assert zero hit when hits is set to False. """

        if method == 'GET':
            res = self.request(
                endpoint,
                params=kwargs,
                data=data,
                json=json).json()

            assert bool(res.get('hits', [])) == hits
            return res

        if method == 'POST':
            res = self.request(
                endpoint,
                method=method,
                params=kwargs,
                data=data,
                json=json).json()

            for item in res:  # list
                if "_id" not in item:
                    _hits = False
                    break
            else:
                _hits = bool(res)
            assert _hits is hits
            return res

        raise ValueError(f'Query method {method} is not supported.')

    @staticmethod
    def msgpack_ok(packed_bytes):
        ''' Load msgpack into a dict '''
        try:
            import msgpack
        except ImportError:
            pytest.skip('Msgpack is not installed.')
        try:
            dic = msgpack.unpackb(packed_bytes)
        except BaseException:  # pylint: disable=bare-except
            assert False, 'Not a valid Msgpack binary.'
        return dic



import os
import time
import math
import copy
from datetime import datetime
import pickle
import asyncio
from functools import partial

from elasticsearch import Elasticsearch

from biothings.utils.hub_db import get_src_build
from biothings.utils.common import timesofar, get_random_string, \
    get_class_from_classpath
from biothings.utils.loggers import get_logger
from biothings.utils.manager import BaseManager
from biothings.utils.es import ESIndexer
from biothings import config as btconfig
from biothings.utils.mongo import doc_feeder, id_feeder
from config import LOG_FOLDER, logger as logging
from biothings.hub.databuild.backend import create_backend, merge_src_build_metadata
from biothings.hub import INDEXER_CATEGORY, INDEXMANAGER_CATEGORY


def new_index_worker(col_name, ids, pindexer, batch_num):
    col = create_backend(col_name).target_collection
    idxer = pindexer()
    cur = doc_feeder(col,
                     step=len(ids),
                     inbatch=False,
                     query={'_id': {
                         '$in': ids
                     }})
    cnt = idxer.index_bulk(cur)
    return cnt


def merge_index_worker(col_name, ids, pindexer, batch_num):
    col = create_backend(col_name).target_collection
    idxer = pindexer()
    upd_cnt = 0
    new_cnt = 0
    cur = doc_feeder(col,
                     step=len(ids),
                     inbatch=False,
                     query={'_id': {
                         '$in': ids
                     }})
    docs = [d for d in cur]
    [d.pop("_timestamp", None) for d in docs]
    # dids = dict([(d["_id"], d) for d in docs])
    dids = {d["_id"]: d for d in docs}
    # dexistings = dict([(d["_id"], d)
    #                    for d in idxer.get_docs([k for k in dids.keys()])])
    dexistings = {d["_id"]: d for d in idxer.get_docs([k for k in dids.keys()])}
    for _id in dexistings:
        d = dexistings[_id]
        # update in-place
        d.update(dids[_id])
        # mark as processed/updated
        dids.pop(_id)
    # updated docs (those existing in col *and* index)
    upd_cnt = idxer.index_bulk(dexistings.values(), len(dexistings))
    logging.debug("%s documents updated in index", repr(upd_cnt))
    # new docs (only in col, *not* in index)
    new_cnt = idxer.index_bulk(dids.values(), len(dids))
    logging.debug("%s new documents in index", repr(new_cnt))
    # need to return one: tuple(cnt,list)
    ret = (upd_cnt[0] + new_cnt[0], upd_cnt[1] + new_cnt[1])
    return ret


def indexer_worker(col_name,
                   ids,
                   pindexer,
                   batch_num,
                   mode="index",
                   worker=new_index_worker):
    try:
        if mode in ["index", "merge"]:
            return worker(col_name, ids, pindexer, batch_num)
        elif mode == "resume":
            idxr = pindexer()
            es_ids = idxr.mexists(ids)
            missing_ids = [e[0] for e in es_ids if e[1] is False]
            if missing_ids:
                return worker(col_name, missing_ids, pindexer, batch_num)
            else:
                # fake indexer results, it has to be a tuple, first elem is num of indexed docs
                return (0, None)
    except Exception as e:
        logger_name = "index_%s_%s_batch_%s" % (pindexer.keywords.get(
            "index", "index"), col_name, batch_num)
        logger, _ = get_logger(logger_name, btconfig.LOG_FOLDER)
        logger.exception("indexer_worker failed")
        exc_fn = os.path.join(btconfig.LOG_FOLDER, "%s.pick" % logger_name)
        pickle.dump({"exc": e, "ids": ids}, open(exc_fn, "wb"))
        logger.info("Exception and IDs were dumped in pickle file '%s'", exc_fn)
        raise


class IndexerException(Exception):
    pass


class Indexer(object):
    """
    Basic indexer, reading documents from a mongo collection (target_name)
    and sending documents to ES.
    """
    def __init__(self, es_host, target_name=None, **kwargs):
        self.host = es_host
        self.env = None
        self.log_folder = LOG_FOLDER
        self.timestamp = datetime.now()
        self.conf_name = None
        self.build_doc = None
        self.target_name = None
        self.index_name = None
        self.doc_type = None
        self.num_shards = None
        self.num_replicas = None
        self.kwargs = kwargs
        self.ti = time.time()

    def get_predicates(self):
        return []

    def get_pinfo(self):
        """
        Return dict containing information about the current process
        (used to report in the hub)
        """
        return {
            "category": INDEXER_CATEGORY,
            "source": "%s:%s" % (self.conf_name, self.index_name),
            "step": "",
            "description": ""
        }
        preds = self.get_predicates()
        if preds:
            pinfo["__predicates__"] = preds
        return pinfo

    @asyncio.coroutine
    def index(self,
              target_name,
              index_name,
              job_manager,
              steps=["index", "post"],
              batch_size=10000,
              ids=None,
              mode="index",
              worker=None):
        """
        Build an index named "index_name" with data from collection "target_collection".

        "ids" can be passed to selectively index documents.

        "mode" can have the following values:
            - 'purge': will delete index if it exists
            - 'resume': will use existing index and add documents. "ids" can be passed as a list of missing IDs,
                    or, if not pass, ES will be queried to identify which IDs are missing for each batch in
                    order to complete the index.
            - 'merge': will merge data with existing index' documents, used when populated several distinct times (cold/hot merge for instance)
            - None (default): will create a new index, assuming it doesn't already exist
        """
        assert job_manager
        # check what to do
        if type(steps) == str:
            steps = [steps]
        self.target_name = target_name
        self.index_name = index_name
        self.load_build()
        self.setup_log()
        # select proper index worker according to mode:
        if worker is None:  # none specified, choose correct one
            if mode == "merge":
                worker = merge_index_worker
            else:
                worker = new_index_worker

        got_error = False
        cnt = 0

        if "index" in steps:
            self.register_status("indexing",
                                 transient=True,
                                 init=True,
                                 job={"step": "index"})
            assert self.build_doc.get("backend_url")
            target_collection = create_backend(
                self.build_doc["backend_url"]).target_collection
            backend_url = self.build_doc["backend_url"]
            _mapping = self.get_mapping()
            _extra = self.get_index_creation_settings()
            # partially instantiated indexer instance for process workers
            partial_idxer = partial(ESIndexer,
                                    doc_type=self.doc_type,
                                    index=index_name,
                                    es_host=self.host,
                                    step=batch_size,
                                    number_of_shards=self.num_shards,
                                    number_of_replicas=self.num_replicas,
                                    **self.kwargs)
            # instantiate one here for index creation
            es_idxer = partial_idxer()
            if es_idxer.exists_index():
                if mode == "purge":
                    es_idxer.delete_index()
                elif mode not in ["resume", "merge"]:
                    msg = "Index already '%s' exists, (use mode='purge' to auto-delete it or mode='resume' to add more documents)" % index_name
                    self.register_status("failed", job={"err": msg})
                    raise IndexerException(msg)

            if mode not in ["resume", "merge"]:
                try:
                    es_idxer.create_index({self.doc_type: _mapping}, _extra)
                except Exception as e:
                    self.logger.exception("Failed to create index")
                    self.register_status("failed", job={"err": repr(e)})
                    raise

            def clean_ids(ids):
                # can't use a generator, it's going to be pickled
                cleaned = []
                for _id in ids:
                    if type(_id) != str:
                        self.logger.warning(
                            "_id '%s' has invalid type (!str), skipped", repr(_id)
                        )
                        continue
                    if len(_id) > 512:  # this is an ES6 limitation
                        self.logger.warning("_id is too long: '%s'", _id)
                        continue
                    cleaned.append(_id)
                return cleaned

            jobs = []
            total = target_collection.count()
            btotal = math.ceil(total / batch_size)
            bnum = 1
            if ids:
                self.logger.info(
                    "Indexing from '%s' with specific list of _ids, create indexer job with batch_size=%d",
                    target_name, batch_size)
                id_provider = iter_n(ids,batch_size)
            else:
                self.logger.info(
                    "Fetch _ids from '%s', and create indexer job with batch_size=%d",
                    target_name, batch_size)
                id_provider = id_feeder(target_collection,
                                        batch_size=batch_size,
                                        logger=self.logger)
            for ids in id_provider:
                yield from asyncio.sleep(0.0)
                origcnt = len(ids)
                ids = clean_ids(ids)
                newcnt = len(ids)
                if origcnt != newcnt:
                    self.logger.warning(
                        "%d document(s) can't be indexed and will be skipped (invalid _id)",
                        origcnt-newcnt)
                # progress count
                cnt += len(ids)
                pinfo = self.get_pinfo()
                pinfo["step"] = self.target_name
                try:
                    descprogress = cnt / total * 100
                except ZeroDivisionError:
                    descprogress = 0.0
                pinfo["description"] = "#%d/%d (%.1f%%)" % (bnum, btotal,
                                                            descprogress)
                self.logger.info("Creating indexer job #%d/%d, to index '%s' %d/%d (%.1f%%)",
                                 bnum, btotal, backend_url, cnt, total, descprogress)
                job = yield from job_manager.defer_to_process(
                    pinfo,
                    partial(indexer_worker, backend_url, ids, partial_idxer,
                            bnum, mode, worker))

                def batch_indexed(f, batch_num):
                    nonlocal got_error
                    try:
                        res = f.result()
                        if type(res) != tuple or type(res[0]) != int:
                            got_error = Exception("Batch #%s failed while indexing collection '%s' [result:%s]" %
                                                  (batch_num, self.target_name, repr(res)))
                    except Exception as e:
                        got_error = e
                        self.logger.exception("Batch indexed error")
                        return

                job.add_done_callback(partial(batch_indexed, batch_num=bnum))
                jobs.append(job)
                bnum += 1
                # raise error as soon as we know
                if got_error:
                    self.register_status("failed",
                                         job={"err": repr(got_error)})
                    raise got_error
            self.logger.info("%d jobs created for indexing step", len(jobs))
            tasks = asyncio.gather(*jobs)

            def done(f):
                nonlocal got_error
                if None in f.result():
                    got_error = None
                    return
                # compute overall inserted/updated records
                # returned values looks like [(num,[]),(num,[]),...]
                cnt = sum((val[0] for val in f.result()))
                self.register_status("success",
                                     job={"step": "index"},
                                     index={"count": cnt})
                if total != cnt:
                    # raise error if counts don't match, but index is still created,
                    # fully registered in case we want to use it anyways
                    err = "Merged collection has %d documents but %d have been indexed (check logs for more)" % (
                        total, cnt)
                    raise IndexerException(err)
                self.logger.info(
                    "Index '%s' successfully created using merged collection %s",
                    index_name, target_name,
                    extra={"notify": True})

            tasks.add_done_callback(done)
            yield from tasks

        if "post" in steps:
            self.logger.info("Running post-index process for index '%s'",
                             index_name)
            self.register_status("indexing",
                                 transient=True,
                                 init=True,
                                 job={"step": "post-index"})
            pinfo = self.get_pinfo()
            pinfo["step"] = "post_index"
            # for some reason (like maintaining object's state between pickling).
            # we can't use process there. Need to use thread to maintain that state without
            # building an unmaintainable monster
            job = yield from job_manager.defer_to_thread(
                pinfo,
                partial(self.post_index,
                        target_name,
                        index_name,
                        job_manager,
                        steps=steps,
                        batch_size=batch_size,
                        ids=ids,
                        mode=mode))

            def posted(f):
                nonlocal got_error
                try:
                    res = f.result()
                    self.logger.info(
                        "Post-index process done for index '%s': %s",
                        index_name, res)
                    self.register_status("indexing",
                                         job={"step": "post-index"})
                except Exception as e:
                    got_error = e
                    self.logger.exception(
                        "Post-index process failed for index '%s':",
                        index_name,
                        extra={"notify": True})
                    return

            job.add_done_callback(posted)
            yield from job  # consume future

        if got_error:
            self.register_status("failed", job={"err": repr(got_error)})
            raise got_error
        else:
            self.register_status("success")
            return {"%s" % self.index_name: cnt}

    def register_status(self, status, transient=False, init=False, **extra):
        assert self.build_doc
        src_build = get_src_build()
        job_info = {
            'status': status,
            'step_started_at': datetime.now(),
            'logfile': self.logfile,
        }
        index_info = {
            "index": {
                self.index_name: {
                    'host': self.host,
                    'environment': self.env,
                    'conf_name': self.conf_name,
                    'target_name': self.target_name,
                    'index_name': self.index_name,
                    'doc_type': self.doc_type,
                    'num_shards': self.num_shards,
                    'num_replicas': self.num_replicas
                }
            }
        }
        if transient:
            # record some "in-progress" information
            job_info['pid'] = os.getpid()
        else:
            # only register time when it's a final state
            job_info["time"] = timesofar(self.ti)
            t1 = round(time.time() - self.ti, 0)
            job_info["time_in_s"] = t1
            index_info["index"][self.index_name]["created_at"] = datetime.now()
        if "index" in extra:
            index_info["index"][self.index_name].update(extra["index"])
        if "job" in extra:
            job_info.update(extra["job"])
        # since the base is the merged collection, we register info there
        build = src_build.find_one({'_id': self.target_name})
        assert build, "Can't find build document '%s'" % self.target_name
        if init:
            # init timer for this step
            self.ti = time.time()
            src_build.update({'_id': self.target_name},
                             {"$push": {
                                 'jobs': job_info
                             }})
            # now refresh/sync
            build = src_build.find_one({'_id': self.target_name})
        else:
            # merge extra at root level
            # (to keep building data...) and update the last one
            # (it's been properly created before when init=True)
            build["jobs"] and build["jobs"][-1].update(job_info)

            def merge_index_info(target, d):
                if "__REPLACE__" in d.keys():
                    d.pop("__REPLACE__")
                    target = d
                else:
                    for k, v in d.items():
                        if type(v) == dict:
                            if k in target:
                                target[k] = merge_index_info(target[k], v)
                            else:
                                v.pop("__REPLACE__", None)
                                # merge v with "nothing" just to make sure to remove any "__REPLACE__"
                                v = merge_index_info({}, v)
                                target[k] = v
                        else:
                            target[k] = v
                return target

            build = merge_index_info(build, index_info)
            src_build.replace_one({"_id": build["_id"]}, build)

    def post_index(self,
                   target_name,
                   index_name,
                   job_manager,
                   steps=["index", "post"],
                   batch_size=10000,
                   ids=None,
                   mode=None):
        """
        Override in sub-class to add a post-index process. Method's signature is the same as index() to get
        the full context. This method will run in a thread (using job_manager.defer_to_thread())
        """
        pass

    def setup_log(self):
        self.logger, self.logfile = get_logger('index_%s' % self.index_name,
                                               self.log_folder)

    def get_index_creation_settings(self):
        """
        Override to return a dict containing some extra settings
        for index creation. Dict will be merged with mandatory settings,
        see biothings.utils.es.ESIndexer.create_index for more.
        """
        return {
            # as of ES6, include_in_all was removed, we need to create our own "all" field
            "query": {
                "default_field": "all"
            },
            "codec": "best_compression",
            # as of ES6, analysers/tokenizers must be defined in index settings, during creation
            "analysis": {
                "analyzer": {
                    # soon deprecated in favor of keyword_lowercase_normalizer
                    "string_lowercase": {
                        "tokenizer": "keyword",
                        "filter": "lowercase"
                    },
                    "whitespace_lowercase": {
                        "tokenizer": "whitespace",
                        "filter": "lowercase"
                    },
                },
                "normalizer": {
                    "keyword_lowercase_normalizer": {
                        "filter": ["lowercase"],
                        "type": "custom",
                        "char_filter": []
                    },
                }
            },
        }

    def enrich_final_mapping(self, final_mapping):
        """
        final_mapping is the ES mapping ready to be sent,
        (with "dynamic" and "all" at its root for instance)
        this method gives opportunity to add more mapping definitions
        not directly related to datasources, such as other root keys
        """
        return final_mapping

    def get_mapping(self):
        '''collect mapping data from data sources.
        '''
        mapping = self.build_doc.get("mapping", {})
        # default "all" field to replace include_in_all field in older versions of ES
        mapping["all"] = {'type': 'text'}
        final_mapping = {"properties": mapping, "dynamic": "false"}
        final_mapping = self.enrich_final_mapping(final_mapping)
        final_mapping["_meta"] = self.get_metadata()

        return final_mapping

    def get_metadata(self):
        return self.build_doc.get("_meta", {})

    def get_build(self, target_name=None):
        target_name = target_name or self.target_name
        assert target_name, "target_name must be defined first before searching for builds"
        builds = [b for b in self.build_config["build"] if b == target_name]
        assert len(
            builds
        ) == 1, "Can't find build for config '%s' and target_name '%s'" % (
            self.conf_name, self.target_name)
        return self.build_config["build"][builds[0]]

    def get_src_versions(self):
        build = self.get_build()
        return build["src_version"]

    def get_stats(self):
        build = self.get_build()
        return build["stats"]

    def get_timestamp(self):
        build = self.get_build()
        return build["build_date"]

    def get_build_version(self):
        build = self.get_build()
        return build["build_version"]

    def load_build(self, target_name=None):
        '''Load build info from src_build collection.'''
        target_name = target_name or self.target_name
        src_build = get_src_build()
        self.build_doc = src_build.find_one({'_id': target_name})
        assert self.build_doc, "Can't find build document associated to '%s'" % target_name
        _cfg = self.build_doc.get("build_config")
        if _cfg:
            self.build_config = _cfg
            #if not "doc_type" in _cfg:
            #    raise ValueError("Missing 'doc_type' in build config")
            self.doc_type = _cfg.get("doc_type")
            self.num_shards = _cfg.get("num_shards", 10)  # optional
            self.num_shards = self.num_shards and int(
                self.num_shards) or self.num_shards
            self.num_replicas = _cfg.get("num_replicas", 0)  # optional
            self.num_replicas = self.num_replicas and int(
                self.num_replicas) or self.num_replicas
            self.conf_name = _cfg["name"]
        else:
            raise ValueError("Cannot find build config associated to '%s'" %
                             target_name)
        return _cfg


class ColdHotIndexer(Indexer):
    """
    This indexer works with 2 mongo collections to create a single index.
    - one premerge collection contains "cold" data, which never changes (not updated)
    - another collection contains "hot" data, regularly updated
    Index is created fetching the premerge documents. Then, documents from the hot collection
    are merged by fetching docs from the index, updating them, and putting them back in the index.
    """
    def __init__(self, *args, **kwargs):
        super(ColdHotIndexer, self).__init__(*args, **kwargs)
        self.hot_target_name = None
        self.cold_target_name = None
        self.cold_build_doc = None
        self.hot_build_doc = None
        self.cold_cfg = None
        self.hot_cfg = None

    @asyncio.coroutine
    def index(self,
              hot_name,
              index_name,
              job_manager,
              steps=["index", "post"],
              batch_size=10000,
              ids=None,
              mode="index"):
        """
        Same as Indexer.index method but works with a cold/hot collections strategy: first index the cold collection then
        complete the index with hot collection (adding docs or merging them in existing docs within the index)
        """
        assert job_manager
        # check what to do
        if type(steps) == str:
            steps = [steps]
        self.hot_target_name = hot_name
        self.setup_log()
        self.load_build()
        if type(index_name) == list:
            # values are coming from target names, use the cold
            self.index_name = self.hot_target_name
        else:
            self.index_name = index_name
        got_error = False
        cnt = 0
        if "index" in steps:
            # selectively index cold then hot collections, using default index method
            # but specifically 'index' step to prevent any post-process before end of
            # index creation
            # Note: copy backend values as there are some references values between cold/hot and build_doc
            hot_backend_url = self.hot_build_doc["backend_url"]
            cold_backend_url = self.cold_build_doc["backend_url"]
            # target collection is taken from backend_url field, temporarily override.
            self.build_doc["backend_url"] = cold_backend_url
            cold_task = super(ColdHotIndexer,
                              self).index(self.cold_target_name,
                                          self.index_name,
                                          steps="index",
                                          job_manager=job_manager,
                                          batch_size=batch_size,
                                          ids=ids,
                                          mode=mode)
            # wait until cold is fully indexed
            yield from cold_task
            # use updating indexer worker for hot to merge in index
            # back to hot collection
            self.build_doc["backend_url"] = hot_backend_url
            hot_task = super(ColdHotIndexer,
                             self).index(self.hot_target_name,
                                         self.index_name,
                                         steps="index",
                                         job_manager=job_manager,
                                         batch_size=batch_size,
                                         ids=ids,
                                         mode="merge")
            task = asyncio.ensure_future(hot_task)

            def done(f):
                nonlocal got_error
                nonlocal cnt
                try:
                    res = f.result()
                    # compute overall inserted/updated records
                    cnt = sum(res.values())
                    self.register_status("success",
                                         job={"step": "index"},
                                         index={"count": cnt})
                    self.logger.info("index '%s' successfully created",
                                     index_name,
                                     extra={"notify": True})
                except Exception as e:
                    logging.exception("failed indexing cold/hot collections:")
                    got_error = e
                    raise

            task.add_done_callback(done)
            yield from task
            if got_error:
                raise got_error
        if "post" in steps:
            # use super index but this time only on hot collection (this is the entry point, cold collection
            # remains hidden from outside)
            hot_task = super(ColdHotIndexer,
                             self).index(self.hot_target_name,
                                         self.index_name,
                                         steps="post",
                                         job_manager=job_manager,
                                         batch_size=batch_size,
                                         ids=ids,
                                         mode=mode)
            task = asyncio.ensure_future(hot_task)

            def posted(f):
                nonlocal got_error
                try:
                    _ = f.result()
                    # no need to process the return value more, it's been done in super
                except Exception as e:
                    self.logger.exception(
                        "Post-index process failed for index '%s':",
                        self.index_name,
                        extra={"notify": True})
                    got_error = e
                    raise

            task.add_done_callback(posted)
            yield from task
            if got_error:
                raise got_error

        return {self.index_name: cnt}

    # by default, build_doc is considered to be the hot one
    # (mainly used so we can call super methods as parent)
    @property
    def build_doc(self):
        return self.hot_build_doc

    @build_doc.setter
    def build_doc(self, val):
        self.hot_build_doc = val

    def get_mapping(self):
        final_mapping = super(ColdHotIndexer, self).get_mapping()
        cold_mapping = self.cold_build_doc.get("mapping", {})
        final_mapping["properties"].update(cold_mapping)  # mix cold&hot
        return final_mapping

    def get_metadata(self):
        meta = merge_src_build_metadata(
            [self.cold_build_doc, self.hot_build_doc])
        return meta

    def get_src_versions(self):
        _meta = self.get_metadata()
        return _meta["src_version"]

    def get_stats(self):
        _meta = self.get_metadata()
        return _meta["stats"]

    def get_timestamp(self):
        _meta = self.get_metadata()
        return _meta["build_date"]

    def get_build_version(self):
        _meta = self.get_metadata()
        return _meta["build_version"]

    def load_build(self):
        """
        Load cold and hot build documents.
        Index settings are the one declared in the hot build doc.
        """
        src_build = get_src_build()
        # we don't want to reload build docs if they are already loaded
        # so we can temporarily override values when dealing with cold/hot collection
        # (kind of a hack, not really clean, but...)
        if self.hot_build_doc and self.cold_build_doc and self.build_doc:
            self.logger.debug("Build documents already loaded")
            return
        self.hot_build_doc = src_build.find_one({'_id': self.hot_target_name})
        # search the cold collection definition
        assert "build_config" in self.hot_build_doc and "cold_collection" in self.hot_build_doc["build_config"], \
            "Can't find cold_collection field in build_config"
        self.cold_target_name = self.hot_build_doc["build_config"][
            "cold_collection"]
        self.cold_build_doc = src_build.find_one(
            {'_id': self.cold_target_name})
        # we'll register everything (status) on the hot one
        self.build_doc = self.hot_build_doc
        assert self.cold_build_doc, "Can't find build document associated to '%s'" % self.cold_target_name
        assert self.hot_build_doc, "Can't find build document associated to '%s'" % self.hot_target_name
        self.cold_cfg = self.cold_build_doc.get("build_config")
        self.hot_cfg = self.hot_build_doc.get("build_config")
        if self.hot_cfg or not self.cold_cfg:
            self.build_config = self.hot_cfg
            if "doc_type" not in self.hot_cfg:
                raise ValueError("Missing 'doc_type' in build config")
            self.doc_type = self.hot_cfg["doc_type"]
            self.num_shards = self.hot_cfg.get("num_shards", 10)  # optional
            self.num_shards = self.num_shards and int(
                self.num_shards) or self.num_shards
            self.num_replicas = self.hot_cfg.get("num_replicas", 0)  # optional
            self.num_replicas = self.num_replicas and int(
                self.num_replicas) or self.num_replicas
            self.conf_name = self.hot_cfg["name"]
        else:
            raise ValueError(
                "Cannot find build config associated to '%s' or '%s'" %
                (self.hot_target_name, self.cold_target_name))
        return (self.cold_cfg, self.hot_cfg)


class IndexManager(BaseManager):

    DEFAULT_INDEXER = Indexer

    def __init__(self, *args, **kwargs):
        super(IndexManager, self).__init__(*args, **kwargs)
        self.src_build = get_src_build()
        self.indexers = {}
        self.es_config = {}
        self.t0 = time.time()
        self.prepared = False
        self.log_folder = LOG_FOLDER
        self.timestamp = datetime.now()
        self.setup()

    def clean_stale_status(self):
        src_build = get_src_build()
        for build in src_build.find():
            dirty = False
            for job in build.get("jobs", []):
                if job.get("status") == "indexing":
                    logging.warning(
                        "Found stale build '%s', marking index status as 'canceled'",
                        build["_id"])
                    job["status"] = "canceled"
                    dirty = True
            if dirty:
                src_build.replace_one({"_id": build["_id"]}, build)

    def setup(self):
        self.setup_log()

    def setup_log(self):
        self.logger, self.logfile = get_logger('indexmanager', self.log_folder)

    def get_predicates(self):
        def no_other_indexmanager_step_running(job_manager):
            """IndexManager deals with snapshot, publishing,
            none of them should run more than one at a time"""
            return len([
                j for j in job_manager.jobs.values()
                if j["category"] == INDEXMANAGER_CATEGORY
            ]) == 0

        return [no_other_indexmanager_step_running]

    def get_pinfo(self):
        """
        Return dict containing information about the current process
        (used to report in the hub)
        """
        pinfo = {
            "category": INDEXMANAGER_CATEGORY,
            "source": "",
            "step": "",
            "description": ""
        }
        preds = self.get_predicates()
        if preds:
            pinfo["__predicates__"] = preds
        return pinfo

    def __getitem__(self, conf_name):
        """
        Return an instance of an indexer for the build configuration named 'conf_name'
        Note: each call returns a different instance (factory call behind the scene...)
        """
        kwargs = BaseManager.__getitem__(self, conf_name)
        return kwargs

    def configure_from_list(self, indexers_kwargs):
        for dindex in indexers_kwargs:
            assert len(
                dindex) == 1, "Invalid indexer registration data: %s" % dindex
            env, idxkwargs = list(dindex.items())[0]
            self.register[env] = idxkwargs

    def configure_from_dict(self, confdict):
        self.es_config = copy.deepcopy(confdict)
        self.indexers.update(confdict.get("indexer_select", {}))
        indexers_kwargs = []
        for env, conf in confdict["env"].items():
            idxkwargs = dict(**conf["indexer"]["args"])
            # propagate ES host to indexer's kwargs
            idxkwargs["es_host"] = self.es_config["env"][env]["host"]
            indexers_kwargs.append({env: idxkwargs})
        self.configure_from_list(indexers_kwargs)

    def configure(self, indexer_defs):
        """
        Register indexers with:
        - a list of dict as:
            [{"indexer_type_name": partial},{....}]
        - a dict containing all indexer definitions:
            {"env" : {
                "env1" : {
                    "host": "localhost:9200",
                    "timeout": ..., "retry":...,
                    "indexer" : "path.to.ClassIndexer",
                },
                ...
            }
        Partial is used to instantiate an indexer, without args
        """
        if type(indexer_defs) == list:
            self.configure_from_list(indexer_defs)
        elif type(indexer_defs) == dict:
            self.configure_from_dict(indexer_defs)
        else:
            raise ValueError(
                "Unknown indexer definitions type (expecting a list or a dict")
        self.logger.info(self.indexers)
        self.logger.info(self.register)

    def find_indexer(self, target_name):
        """
        Return indexer class required to index target_name.
        Rules depend on what's inside the corresponding src_build doc
        and the indexers definitions
        """
        doc = self.src_build.find_one({"_id": target_name})
        if not self.indexers or not doc:
            return self.__class__.DEFAULT_INDEXER
        klass = None
        for path_in_doc in self.indexers:
            if klass is None and (
                    path_in_doc is None
                    or path_in_doc == "default"
                    or path_in_doc == ""):
                # couldn't find a klass yet and we found a default declated, keep it
                strklass = self.indexers[path_in_doc]
                klass = get_class_from_classpath(strklass)
            else:
                try:
                    strklass = self.indexers[path_in_doc]
                    klass = get_class_from_classpath(strklass)
                    self.logger.info(
                        "Found special indexer '%s' required to index '%s'",
                        klass, target_name)
                    # the first to match wins
                    break
                except KeyError:
                    pass
        if klass is None:
            self.logger.debug("Using default indexer")
            return self.__class__.DEFAULT_INDEXER
        else:
            # either we return a default declared in config or
            # a specific one found according to the doc
            self.logger.debug("Using custom indexer %s", klass)
            return klass

    def index(self,
              indexer_env,
              target_name=None,
              index_name=None,
              ids=None,
              **kwargs):
        """
        Trigger an index creation to index the collection target_name and create an
        index named index_name (or target_name if None). Optional list of IDs can be
        passed to index specific documents.
        """

        def indexed(f):
            try:
                res = f.result()
                self.logger.info(
                    "Done indexing target '%s' to index '%s': %s",
                    target_name, index_name, res)
            except Exception:
                self.logger.exception("Error while running index job:")
                raise

        idxklass = self.find_indexer(target_name)
        idxkwargs = self[indexer_env]
        idx = idxklass(**idxkwargs)
        idx.env = indexer_env
        idx.target_name = target_name
        index_name = index_name or target_name
        job = idx.index(target_name,
                        index_name,
                        ids=ids,
                        job_manager=self.job_manager,
                        **kwargs)
        job = asyncio.ensure_future(job)
        job.add_done_callback(indexed)

        return job

    def update_metadata(self,
                        indexer_env,
                        index_name,
                        build_name=None,
                        _meta=None):
        """
        Update _meta for index_name, based on build_name (_meta directly
        taken from the src_build document) or _meta
        """
        idxkwargs = self[indexer_env]
        # 1st pass we get the doc_type (don't want to ask that on the signature...)
        indexer = create_backend(
            (idxkwargs["es_host"], index_name, None)).target_esidxer
        m = indexer._es.indices.get_mapping(index_name)
        assert len(m[index_name]["mappings"]) == 1, "Found more than one doc_type: " + \
            "%s" % m[index_name]["mappings"].keys()
        doc_type = list(m[index_name]["mappings"].keys())[0]
        # 2nd pass to re-create correct indexer
        indexer = create_backend(
            (idxkwargs["es_host"], index_name, doc_type)).target_esidxer
        if build_name:
            build = get_src_build().find_one({"_id": build_name})
            assert build, "No such build named '%s'" % build_name
            _meta = build.get("_meta")
        assert _meta is not None, "No _meta found"
        return indexer.update_mapping_meta({"_meta": _meta})

    def index_info(self, env=None, remote=False):
        res = copy.deepcopy(self.es_config)
        for kenv in self.es_config["env"]:
            if env and env != kenv:
                continue
            if remote:
                # lost all indices, remotely
                try:
                    cl = Elasticsearch(res["env"][kenv]["host"],
                                       timeout=1,
                                       max_retries=0)
                    indices = [{
                        "index": k,
                        "doc_type": list(v["mappings"].keys())[0],
                        "aliases": list(v["aliases"].keys())
                    } for k, v in cl.indices.get("*").items()]
                    # init index key if not done in config var
                    # (a default index name can be specified in config...)
                    if "index" not in res["env"][kenv]:
                        res["env"][kenv]["index"] = []
                    # for now, we just consider
                    if type(res["env"][kenv]["index"]) == dict:
                        # we don't where to put those indices because we don't
                        # have that information, so we just put those in a default category
                        # TODO: put that info in metadata ?
                        res["env"][kenv]["index"].setdefault(
                            None, []).extend(indices)
                    else:
                        assert type(res["env"][kenv]["index"]) == list
                        res["env"][kenv]["index"].extend(indices)
                except Exception:
                    self.logger.exception("Can't load remote indices:")
                    continue
        return res

    def validate_mapping(self, mapping, env):
        idxkwargs = self[env]
        # just get the default indexer (target_name doesn't exist, return default one)
        idxklass = self.find_indexer(target_name="__placeholder_name__%s" %
                                     get_random_string())
        idxr_obj = idxklass(**idxkwargs)
        settings = idxr_obj.get_index_creation_settings()
        # generate a random index, it'll be deleted at the end
        index_name = ("hub_tmp_%s" % get_random_string()).lower()
        idxr = ESIndexer(index=index_name,
                         es_host=idxr_obj.host,
                         doc_type=None)
        self.logger.info("Testing mapping by creating index '%s' on host '%s' (settings: %s)",
                         index_name, idxr_obj.host, settings)
        try:
            res = idxr.create_index(mapping, settings)
            return res
        except Exception as e:
            self.logger.exception("create_index failed")
            raise e
        finally:
            try:
                idxr.delete_index()
            except Exception:
                pass


class DynamicIndexerFactory(object):
    """
    In the context of autohub/standalone instances, create indexer
    with parameters taken from versions.json URL.
    A list of  URLs is provided so the factory knows how to create these
    indexers for each URLs. There's no way to "guess" an ES host from a URL,
    so this parameter must be specified as well, common to all URLs
    "suffix" param is added at the end of index names.
    """

    def __init__(self, urls, es_host, suffix="_current"):
        self.urls = urls
        self.es_host = es_host
        self.bynames = {}
        for url in urls:
            if isinstance(url, dict):
                name = url["name"]
                # actual_url = url["url"]
            else:
                name = os.path.basename(os.path.dirname(url))
                # actual_url = url
            self.bynames[name] = {
                "es_host": self.es_host,
                "index": name + suffix
            }

    def create(self, name):
        conf = self.bynames[name]
        pidxr = partial(ESIndexer, index=conf["index"],
                        doc_type=None,
                        es_host=conf["es_host"])
        conf = {"es_host": conf["es_host"], "index": conf["index"]}
        return pidxr, conf



import sys
import os
import json
import time
import copy
import pickle
from datetime import datetime
from pprint import pformat
import asyncio
from functools import partial
from elasticsearch.helpers import BulkIndexError
from elasticsearch.exceptions import ConflictError

from biothings.utils.common import timesofar, iter_n, loadobj
from biothings.utils.mongo import doc_feeder, get_target_db, invalidate_cache
from biothings.utils.hub_db import get_src_build
from biothings.utils.loggers import get_logger
from biothings import config as btconfig
from biothings.utils.manager import BaseManager
from .backend import create_backend, generate_folder
from ..dataload.storage import UpsertStorage
import biothings.utils.jsonpatch as jsonpatch
from biothings.hub import SYNCER_CATEGORY

logging = btconfig.logger


class SyncerException(Exception):
    pass


class BaseSyncer(object):

    # diff type name, identifying the diff algorithm
    # must be set in sub-class
    diff_type = None

    # backend used to sync data (mongo / es)
    # must be set in sub-class
    target_backend_type = None

    def __init__(self, job_manager, log_folder):
        self.log_folder = log_folder
        self.job_manager = job_manager
        self.timestamp = datetime.now()
        self.ti = time.time()
        self.synced_cols = None  # str representation of synced cols (internal usage)
        self.setup_log()
        # set by manager during instanciation
        self.old = None
        self.new = None
        self.target_backend = None
        self._meta = None

    def setup_log(self):
        self.logger, self.logfile = get_logger('sync')

    def get_predicates(self):
        #def no_same_syncer_running(job_manager):
        #    """Avoid syncers collision"""
        #    return len([j for j in job_manager.jobs.values() if \
        #            j["source"] == self.synced_cols and j["category"] == SYNCER_CATEGORY]) == 0
        return []

    def get_pinfo(self):
        pinfo = {"category": SYNCER_CATEGORY, "step": "", "description": ""}
        preds = self.get_predicates()
        if preds:
            pinfo["__predicates__"] = preds
        return pinfo

    def register_status(self, status, transient=False, init=False, **extra):
        src_build = get_src_build()
        job_info = {
            'status': status,
            'step_started_at': datetime.now(),
            'logfile': self.logfile,
        }
        # to select correct diff sub-record (1 collection can be diffed with multiple others)
        diff_key = "%s" % self.old.target_name
        # once in diff, select correct sync sub-record (1 diff can be applied to different backend)
        # replace dots as hostname can have dots which could be interpreted as dotted field by mongo
        # also remove doc_type (which can be sometimes None if hub deals with multiple APIs,
        # and is not useful in distinguishing where the diff was applid since there's only one
        # doc type allowed now since ES6 (last element in self.target_backend is doc_type)
        sync_key = "-".join(self.target_backend[:-1]).replace(".", "-")
        sync_info = {sync_key: {}}
        if transient:
            # record some "in-progress" information
            job_info['pid'] = os.getpid()
        else:
            # only register time when it's a final state
            job_info["time"] = timesofar(self.ti)
            t1 = round(time.time() - self.ti, 0)
            job_info["time_in_s"] = t1
            sync_info[sync_key]["created_at"] = datetime.now()
        if "sync" in extra:
            sync_info[sync_key].update(extra["sync"])
        if "job" in extra:
            job_info.update(extra["job"])
        # since the base is the merged collection, we register info there
        # as the new collection (diff results are associated to the most recent colleciton)
        build = src_build.find_one({'_id': self.new.target_name})
        if not build:
            self.logger.info(
                "Can't find build document '%s', no status to register" %
                self.new.target_name)
            return
        assert "diff" in build and diff_key in build[
            "diff"], "Missing previous diff information in build document"
        if init:
            # init timer for this step
            self.ti = time.time()
            src_build.update({'_id': self.new.target_name},
                             {"$push": {
                                 'jobs': job_info
                             }})
            # now refresh/sync
            build = src_build.find_one({'_id': self.new.target_name})
        else:
            # merge extra at root level
            # (to keep building data...) and update the last one
            # (it's been properly created before when init=True)
            build["jobs"] and build["jobs"][-1].update(job_info)

            def merge_info(target, d):
                if "__REPLACE__" in d.keys():
                    d.pop("__REPLACE__")
                    target = d
                else:
                    for k, v in d.items():
                        if type(v) == dict:
                            if k in target:
                                target[k] = merge_info(target[k], v)
                            else:
                                v.pop("__REPLACE__", None)
                                # merge v with "nothing" just to make sure to remove any "__REPLACE__"
                                v = merge_info({}, v)
                                target[k] = v
                        else:
                            target[k] = v
                return target

            sync_info = {
                "sync":
                merge_info(build["diff"][diff_key].get("sync", {}), sync_info)
            }
            build["diff"][diff_key].update(sync_info)
            #src_build.update({'_id': build["_id"]}, {"$set": index_info})
            src_build.replace_one({"_id": build["_id"]}, build)

    def load_metadata(self, diff_folder):
        self._meta = json.load(open(os.path.join(diff_folder,
                                                 "metadata.json")))

    def get_target_backend(self):
        # first try to use what's been passed explicitely
        # then default to what's in config (tuple will be used for create_backend() call)
        # or use what we have in the diff metadata
        old_db_col_names = self.target_backend or \
            (btconfig.ES_HOST, btconfig.ES_INDEX_NAME, btconfig.ES_DOC_TYPE) or \
            self._meta["old"]["backend"]
        return old_db_col_names

    @asyncio.coroutine
    def sync_cols(self,
                  diff_folder,
                  batch_size=10000,
                  mode=None,
                  force=False,
                  target_backend=None,
                  steps=["mapping", "content", "meta", "post"],
                  debug=False):
        """
        Sync a collection with diff files located in diff_folder. This folder contains a metadata.json file which
        describes the different involved collection: "old" is the collection/index to be synced, "new" is the collecion
        that should be obtained once all diff files are applied (not used, just informative).
        If target_backend (bt.databbuild.backend.create_backend() notation), then it will replace "old" (that is, the one
        being synced)
        """
        if type(steps) == str:
            steps = [steps]
        assert self.old and self.new, "'self.old' and 'self.new' must be set to old/new collections"
        self.target_backend = target_backend
        got_error = False
        cnt = 0
        jobs = []
        self.load_metadata(diff_folder)
        meta = self._meta
        diff_type = self.diff_type
        selfcontained = "selfcontained" in self._meta["diff"]["type"]
        old_db_col_names = self.get_target_backend()
        new_db_col_names = self._meta["new"]["backend"]
        diff_mapping_file = self._meta["diff"]["mapping_file"]
        pinfo = self.get_pinfo()
        self.synced_cols = "%s -> %s" % (old_db_col_names, new_db_col_names)
        pinfo["source"] = self.synced_cols
        summary = {}
        if "mapping" in steps and self.target_backend_type == "es":
            if diff_mapping_file:
                # old_db_col_names is actually the index name in that case
                index_name = old_db_col_names[1]
                doc_type = self._meta["build_config"]["doc_type"]
                indexer = create_backend(old_db_col_names).target_esidxer
                pinfo["step"] = "mapping"
                pinfo["description"] = diff_mapping_file

                def update_mapping():
                    diffm = os.path.join(diff_folder,
                                         diff_mapping_file["name"])
                    ops = loadobj(diffm)
                    mapping = indexer.get_mapping()
                    # we should have the same doc type declared in the mapping
                    mapping[doc_type]["properties"] = jsonpatch.apply_patch(
                        mapping[doc_type]["properties"], ops)
                    res = indexer.update_mapping(mapping)
                    return res

                self.register_status("syncing",
                                     transient=True,
                                     init=True,
                                     job={"step": "sync-mapping"})
                job = yield from self.job_manager.defer_to_thread(
                    pinfo, partial(update_mapping))

                def updated(f):
                    try:
                        _ = f.result()
                        self.logger.info("Mapping updated on index '%s'" %
                                         index_name)
                        summary["mapping_updated"] = True
                        self.register_status("success",
                                             job={"step": "sync-mapping"},
                                             sync=summary)
                    except Exception as e:
                        nonlocal got_error
                        self.logger.error(
                            "Failed to update mapping on index '%s': %s" %
                            (index_name, e))
                        self.register_status("failed", job={"err": repr(e)})
                        got_error = e

                job.add_done_callback(updated)
                yield from job

            if got_error:
                self.logger.error("Failed to update mapping on index '%s': %s" %
                                  (old_db_col_names, got_error), extra={"notify": True})
                raise got_error

        if "content" in steps:
            if selfcontained:
                # selfconained is a worker param, isolate diff format
                diff_type = diff_type.replace("-selfcontained",
                                              "").replace("-", "_")
            diff_files = [(os.path.join(diff_folder,
                                        e["name"]), e.get("worker_args", {}))
                          for e in self._meta["diff"]["files"]]
            total = len(diff_files)
            self.logger.info("Syncing %s to %s using diff files in '%s'" %
                             (old_db_col_names, new_db_col_names, diff_folder))
            pinfo["step"] = "content"
            self.register_status("syncing",
                                 transient=True,
                                 init=True,
                                 job={"step": "sync-content"})
            for diff_file, worker_args in diff_files:
                cnt += 1
                pinfo["description"] = "file %s (%s/%s)" % (diff_file, cnt,
                                                            total)
                worker = getattr(sys.modules["biothings.hub.databuild.syncer"], "sync_%s_%s_worker" %
                                 (self.target_backend_type, diff_type))
                strwargs = worker_args and " using specific worker args %s" % repr(
                    worker_args) or ""
                self.logger.info(
                    "Creating sync worker %s for file %s (%s/%s)%s" %
                    (worker.__name__, diff_file, cnt, total, strwargs))
                # deepcopy to make we don't embed "self" with unpickleable stuff
                meta = copy.deepcopy(self._meta)
                job = yield from self.job_manager.defer_to_process(
                    pinfo,
                    partial(worker, diff_file, old_db_col_names,
                            new_db_col_names,
                            worker_args.get("batch_size") or batch_size, cnt,
                            force, selfcontained, meta, debug))
                jobs.append(job)

            def synced(f):
                try:
                    res = f.result()
                    for d in res:
                        for k in d:
                            summary.setdefault(k, 0)
                            summary[k] += d[k]
                except Exception as e:
                    nonlocal got_error
                    got_error = e
                    self.register_status("failed", job={"err": repr(e)})
                    raise

            tasks = asyncio.gather(*jobs)
            tasks.add_done_callback(synced)
            yield from tasks
            if got_error:
                self.logger.error("Failed to sync collection from %s to %s using diff files in '%s': %s" %
                                  (old_db_col_names, new_db_col_names, diff_folder, got_error), extra={"notify": True})
                raise got_error
            self.register_status("success",
                                 job={"step": "sync-content"},
                                 sync=summary)

        if "meta" in steps and self.target_backend_type == "es":
            # old_db_col_names is actually the index name in that case
            index_name = old_db_col_names[1]
            doc_type = self._meta["build_config"]["doc_type"]
            indexer = create_backend(old_db_col_names).target_esidxer
            new_meta = self._meta["_meta"]
            pinfo["step"] = "metadata"

            def update_metadata():
                res = indexer.update_mapping_meta({"_meta": new_meta})
                return res

            job = yield from self.job_manager.defer_to_thread(
                pinfo, partial(update_metadata))

            def updated(f):
                try:
                    res = f.result()
                    self.logger.info("Metadata updated on index '%s': %s" %
                                     (index_name, res))
                    summary["metadata_updated"] = True
                    self.register_status("success",
                                         job={"step": "sync-meta"},
                                         sync=summary)
                except Exception as e:
                    nonlocal got_error
                    self.logger.error(
                        "Failed to update metadata on index '%s': %s" %
                        (index_name, e))
                    self.register_status("failed", job={"err": repr(e)})
                    got_error = e

            self.register_status("syncing",
                                 transient=True,
                                 init=True,
                                 job={"step": "sync-meta"})
            job.add_done_callback(updated)
            yield from job

            if got_error:
                self.logger.error("Failed to update metadata on index '%s': %s" %
                                  (old_db_col_names, got_error), extra={"notify": True})
                raise got_error

        if "post" in steps:
            pinfo["step"] = "post"
            job = yield from self.job_manager.defer_to_thread(
                pinfo,
                partial(self.post_sync_cols,
                        diff_folder=diff_folder,
                        batch_size=batch_size,
                        mode=mode,
                        force=force,
                        target_backend=target_backend,
                        steps=steps))

            def posted(f):
                try:
                    res = f.result()
                    self.logger.info(
                        "Post-sync process done on index '%s': %s" %
                        (repr(old_db_col_names), res))
                    summary["post-sync"] = True
                    self.register_status("success",
                                         job={"step": "sync-post"},
                                         sync=summary)
                except Exception as e:
                    nonlocal got_error
                    self.logger.error(
                        "Failed to run post-sync process on index '%s': %s" %
                        (repr(old_db_col_names), e))
                    self.register_status("failed", job={"err": repr(e)})
                    got_error = e

            self.register_status("syncing",
                                 transient=True,
                                 init=True,
                                 job={"step": "sync-post"})
            job.add_done_callback(posted)
            yield from job

            if got_error:
                self.logger.error("Failed to run post-sync process on index '%s': %s" %
                                  (repr(old_db_col_names), got_error), extra={"notify": True})
                raise got_error

        self.logger.info("Succesfully synced index %s to reach collection %s using diff files in '%s': %s" %
                         (old_db_col_names, new_db_col_names, diff_folder, summary), extra={"notify": True})

        return summary

    def post_sync_cols(self, diff_folder, batch_size, mode, force,
                       target_backend, steps):
        """Post-sync hook, can be implemented in sub-class"""
        return

    def sync(self,
             diff_folder=None,
             batch_size=10000,
             mode=None,
             target_backend=None,
             steps=["mapping", "content", "meta", "post"],
             debug=False):
        """wrapper over sync_cols() coroutine, return a task"""
        job = asyncio.ensure_future(
            self.sync_cols(diff_folder=diff_folder,
                           batch_size=batch_size,
                           mode=mode,
                           target_backend=target_backend,
                           steps=steps,
                           debug=debug))
        return job


class ThrottlerSyncer(BaseSyncer):
    def __init__(self, max_sync_workers, *args, **kwargs):
        super(ThrottlerSyncer, self).__init__(*args, **kwargs)
        self.max_sync_workers = max_sync_workers

    def get_predicates(self):
        preds = super(ThrottlerSyncer, self).get_predicates()
        if preds is None:
            preds = []

        def not_too_much_syncers(job_manager):
            """
            Limit number of syncers accordingly (this is useful when live-updating
            the prod,we usually need to reduce the number of sync workers as they
            would kill the ES server otherwise... (or at least produces timeout errors)
            """
            return len([j for j in job_manager.jobs.values() if
                       j["category"] == SYNCER_CATEGORY]) < self.max_sync_workers

        preds.append(not_too_much_syncers)
        return preds


class MongoJsonDiffSyncer(BaseSyncer):
    diff_type = "jsondiff"
    target_backend_type = "mongo"


class MongoJsonDiffSelfContainedSyncer(BaseSyncer):
    diff_type = "jsondiff-selfcontained"
    target_backend_type = "mongo"


class ESJsonDiffSyncer(BaseSyncer):
    diff_type = "jsondiff"
    target_backend_type = "es"


class ESJsonDiffSelfContainedSyncer(BaseSyncer):
    diff_type = "jsondiff-selfcontained"
    target_backend_type = "es"


class ESColdHotJsonDiffSyncer(BaseSyncer):
    diff_type = "coldhot-jsondiff"
    target_backend_type = "es"


class ESColdHotJsonDiffSelfContainedSyncer(BaseSyncer):
    diff_type = "coldhot-jsondiff-selfcontained"
    target_backend_type = "es"


class ThrottledESJsonDiffSyncer(ThrottlerSyncer, ESJsonDiffSyncer):
    pass


class ThrottledESJsonDiffSelfContainedSyncer(ThrottlerSyncer,
                                             ESJsonDiffSelfContainedSyncer):
    pass


class ThrottledESColdHotJsonDiffSyncer(ThrottlerSyncer,
                                       ESColdHotJsonDiffSyncer):
    pass


class ThrottledESColdHotJsonDiffSelfContainedSyncer(
        ThrottlerSyncer, ESColdHotJsonDiffSelfContainedSyncer):
    pass


# TODO: refactor workers (see sync_es_...)
def sync_mongo_jsondiff_worker(diff_file,
                               old_db_col_names,
                               new_db_col_names,
                               batch_size,
                               cnt,
                               force=False,
                               selfcontained=False,
                               metadata={},
                               debug=False):
    """Worker to sync data between a new and an old mongo collection"""
    # check if diff files was already synced
    res = {"added": 0, "updated": 0, "deleted": 0, "skipped": 0}
    synced_file = "%s.synced" % diff_file
    if os.path.exists(synced_file):
        logging.info("Diff file '%s' already synced, skip it" %
                     os.path.basename(diff_file))
        diff = loadobj(synced_file)
        res["skipped"] += len(diff["add"]) + len(diff["delete"]) + len(
            diff["update"])
        return res
    new = create_backend(new_db_col_names)
    old = create_backend(old_db_col_names)
    storage = UpsertStorage(get_target_db(), old.target_collection.name,
                            logging)
    diff = loadobj(diff_file)
    assert new.target_collection.name == diff[
        "source"], "Source is different in diff file '%s': %s" % (
            diff_file, diff["source"])

    # add: get ids from "new"
    if selfcontained:
        # diff["add"] contains all documents, not mongo needed
        for docs in iter_n(diff["add"], batch_size):
            res["added"] += storage.process((d for d in docs), batch_size)
    else:
        cur = doc_feeder(new.target_collection,
                         step=batch_size,
                         inbatch=False,
                         query={'_id': {
                             '$in': diff["add"]
                         }})
        for docs in iter_n(cur, batch_size):
            # use generator otherwise process/doc_iterator will require a dict (that's bad...)
            res["added"] += storage.process((d for d in docs), batch_size)

    # update: get doc from "old" and apply diff
    batch = []
    for patch_info in diff["update"]:
        doc = old.get_from_id(patch_info["_id"])
        try:
            doc = jsonpatch.apply_patch(doc, patch_info["patch"])
            batch.append(doc)
        except jsonpatch.JsonPatchConflict:
            # assuming already applieda
            res["skipped"] += 1
            continue
        if len(batch) >= batch_size:
            res["updated"] += storage.process((d for d in batch), batch_size)
            batch = []
    if batch:
        res["updated"] += storage.process((d for d in batch), batch_size)

    # delete: remove from "old"
    for ids in iter_n(diff["delete"], batch_size):
        res["deleted"] += old.remove_from_ids(ids)

    # we potentially modified the "old" collection so invalidate cache just to make sure
    invalidate_cache(old.target_collection.name, "target")
    logging.info("Done applying diff from file '%s': %s" % (diff_file, res))
    # mark as synced
    os.rename(diff_file, synced_file)
    return res


def sync_es_jsondiff_worker(diff_file,
                            es_config,
                            new_db_col_names,
                            batch_size,
                            cnt,
                            force=False,
                            selfcontained=False,
                            metadata={},
                            debug=False):
    """Worker to sync data between a new mongo collection and an elasticsearch index"""
    res = {"added": 0, "updated": 0, "deleted": 0, "skipped": 0}
    # check if diff files was already synced
    synced_file = "%s.synced" % diff_file
    if os.path.exists(synced_file):
        logging.info("Diff file '%s' already synced, skip it" %
                     os.path.basename(diff_file))
        diff = loadobj(synced_file)
        res["skipped"] += len(diff["add"]) + len(diff["delete"]) + len(
            diff["update"])
        return res
    eskwargs = {}
    # pass optional ES Indexer args
    if hasattr(btconfig, "ES_TIMEOUT"):
        eskwargs["timeout"] = btconfig.ES_TIMEOUT
    if hasattr(btconfig, "ES_MAX_RETRY"):
        eskwargs["max_retries"] = btconfig.ES_MAX_RETRY
    if hasattr(btconfig, "ES_RETRY"):
        eskwargs["retry_on_timeout"] = btconfig.ES_RETRY
    logging.debug("Create ES backend with args: (%s,%s)" %
                  (es_config, eskwargs))
    bckend = create_backend(es_config, **eskwargs)
    indexer = bckend.target_esidxer
    diff = loadobj(diff_file)
    errors = []
    # add: get ids from "new"
    if selfcontained:
        # diff["add"] contains all documents, no mongo needed
        cur = diff["add"]
    else:
        new = create_backend(new_db_col_names)  # mongo collection to sync from
        assert new.target_collection.name == diff[
            "source"], "Source is different in diff file '%s': %s" % (
                diff_file, diff["source"])
        cur = doc_feeder(new.target_collection,
                         step=batch_size,
                         inbatch=False,
                         query={'_id': {
                             '$in': diff["add"]
                         }})
    for docs in iter_n(cur, batch_size):
        # remove potenial existing _timestamp from document
        # (not allowed within an ES document (_source))
        [d.pop("_timestamp", None) for d in docs]
        try:
            res["added"] += indexer.index_bulk(docs,
                                               batch_size,
                                               action="create")[0]
        except BulkIndexError:
            for doc in docs:
                _id = doc.pop("_id")
                try:
                    # force action=create to spot docs already added
                    indexer.index(doc, _id, action="create")
                    res["added"] += 1
                except ConflictError:
                    # already added
                    logging.warning("_id '%s' already added" % _id)
                    res["skipped"] += 1
                    continue
                except Exception as e:
                    errors.append({"_id": _id, "file": diff_file, "error": e})
                    import pickle
                    pickle.dump(errors, open("errors", "wb"))
                    raise
        except Exception as e:
            if debug:
                logging.error(
                    "From diff file '%s', following IDs couldn't be synced because: %s\n%s"
                    % (diff_file, e, [d.get("_id") for d in docs]))
                pickfile = "batch_%s_%s.pickle" % (cnt,
                                                   os.path.basename(diff_file))
                logging.error("Documents pickled in '%s'" % pickfile)
                pickle.dump(docs, open(pickfile, "wb"))
            raise

    # update: get doc from indexer and apply diff
    sync_es_for_update(diff_file, indexer, diff["update"], batch_size, res,
                       debug)

    # delete: remove from "old"
    for ids in iter_n(diff["delete"], batch_size):
        del_skip = indexer.delete_docs(ids)
        res["deleted"] += del_skip[0]
        res["skipped"] += del_skip[1]

    logging.info("Done applying diff from file '%s': %s" % (diff_file, res))
    # mark as synced
    os.rename(diff_file, synced_file)
    return res


def sync_es_coldhot_jsondiff_worker(diff_file,
                                    es_config,
                                    new_db_col_names,
                                    batch_size,
                                    cnt,
                                    force=False,
                                    selfcontained=False,
                                    metadata={},
                                    debug=False):
    res = {"added": 0, "updated": 0, "deleted": 0, "skipped": 0}
    # check if diff files was already synced
    synced_file = "%s.synced" % diff_file
    if os.path.exists(synced_file):
        logging.info("Diff file '%s' already synced, skip it" %
                     os.path.basename(diff_file))
        diff = loadobj(synced_file)
        res["skipped"] += len(diff["add"]) + len(diff["delete"]) + len(
            diff["update"])
        return res
    eskwargs = {}
    # pass optional ES Indexer args
    if hasattr(btconfig, "ES_TIMEOUT"):
        eskwargs["timeout"] = btconfig.ES_TIMEOUT
    if hasattr(btconfig, "ES_MAX_RETRY"):
        eskwargs["max_retries"] = btconfig.ES_MAX_RETRY
    if hasattr(btconfig, "ES_RETRY"):
        eskwargs["retry_on_timeout"] = btconfig.ES_RETRY
    logging.debug("Create ES backend with args: (%s,%s)" %
                  (es_config, eskwargs))
    bckend = create_backend(es_config, **eskwargs)
    indexer = bckend.target_esidxer
    diff = loadobj(diff_file)

    # add: diff between hot collections showed we have new documents but it's
    # possible some of those docs already exist in premerge/cold collection.
    # if so, they should be treated as dict.update() where the hot document content
    # has precedence over the cold content for fields in common
    if selfcontained:
        # diff["add"] contains all documents, no mongo needed
        cur = diff["add"]
    else:
        new = create_backend(new_db_col_names)  # mongo collection to sync from
        assert new.target_collection.name == diff[
            "source"], "Source is different in diff file '%s': %s" % (
                diff_file, diff["source"])
        cur = doc_feeder(new.target_collection,
                         step=batch_size,
                         inbatch=False,
                         query={'_id': {
                             '$in': diff["add"]
                         }})
    for docs in iter_n(cur, batch_size):
        # remove potenial existing _timestamp from document
        # (not allowed within an ES document (_source))
        [d.pop("_timestamp", None) for d in docs]
        # check which docs already exist in existing index (meaning they exist in cold collection)
        dids = dict([(d["_id"], d) for d in docs])
        dexistings = dict([
            (d["_id"], d) for d in indexer.get_docs([k for k in dids.keys()])
        ])
        logging.debug("From current batch, %d already exist" % len(dexistings))
        # remove existing docs from "add" so the rest of the dict will be treated
        # as "real" added documents while update existing ones with new content
        toremove = []
        for _id, d in dexistings.items():
            # update in-place
            if d == dids[d["_id"]]:
                logging.debug("%s was already added, skip it" % d["_id"])
                toremove.append(d["_id"])
                res["skipped"] += 1
            else:
                newd = copy.deepcopy(d)
                d.update(dids[d["_id"]])
                if d == newd:
                    logging.debug("%s was already updated, skip it" % d["_id"])
                    toremove.append(d["_id"])
                    res["skipped"] += 1
            dids.pop(d["_id"])
        for _id in toremove:
            dexistings.pop(_id)
        logging.info("Syncing 'add' documents (%s in total) from cold/hot merge: " % len(docs)
                     + "%d documents will be updated as they already exist in the index, " % len(dexistings)
                     + "%d documents will be added (%d skipped as already processed)" % (len(dids), len(toremove)))
        # treat real "added" documents
        # Note: no need to check for "already exists" errors, as we already checked that before
        # in order to know what to do
        try:
            res["added"] += indexer.index_bulk(dids.values(),
                                               batch_size,
                                               action="create")[0]
        except BulkIndexError:
            logging.error("Error while adding documents %s" %
                          [k for k in dids.keys()])
        # update already existing docs in cold collection
        # treat real "added" documents
        try:
            res["updated"] += indexer.index_bulk(dexistings.values(),
                                                 batch_size)[0]
        except BulkIndexError as e:
            logging.error(
                "Error while updating (via new hot detected docs) documents: %s"
                % e)

    # update: get doc from indexer and apply diff
    # note: it's the same process as for non-coldhot
    sync_es_for_update(diff_file, indexer, diff["update"], batch_size, res,
                       debug)

    # delete: remove from "old"
    for ids in iter_n(diff["delete"], batch_size):
        del_skip = indexer.delete_docs(ids)
        res["deleted"] += del_skip[0]
        res["skipped"] += del_skip[1]

    logging.info("Done applying diff from file '%s': %s" % (diff_file, res))
    # mark as synced
    os.rename(diff_file, synced_file)
    return res


def sync_es_for_update(diff_file, indexer, diffupdates, batch_size, res,
                       debug):
    batch = []
    ids = [p["_id"] for p in diffupdates]
    iterids_bcnt = iter_n(ids, batch_size, True)
    for batchids, bcnt in iterids_bcnt:
        try:
            for i, doc in enumerate(indexer.get_docs(batchids)):
                # recompute correct index in diff["update"], since we split it in batches
                diffidx = i + bcnt - len(
                    batchids
                )  # len(batchids) is not == batch_size for the last one...
                try:
                    patch_info = diffupdates[
                        diffidx]  # same order as what's return by get_doc()...
                    assert patch_info["_id"] == doc["_id"], "%s != %s" % (
                        patch_info["_id"], doc["_id"]
                    )  # ... but just make sure
                    newdoc = jsonpatch.apply_patch(doc, patch_info["patch"])
                    if newdoc == doc:
                        # already applied
                        logging.warning("_id '%s' already synced" % doc["_id"])
                        res["skipped"] += 1
                        continue
                    batch.append(newdoc)
                except jsonpatch.JsonPatchConflict as e:
                    # assuming already applied
                    logging.warning(
                        "_id '%s' already synced ? JsonPatchError: %s" %
                        (doc["_id"], e))
                    res["skipped"] += 1
                    continue
                if len(batch) >= batch_size:
                    res["updated"] += indexer.index_bulk(batch, batch_size)[0]
                    batch = []
            if batch:
                res["updated"] += indexer.index_bulk(batch, batch_size)[0]
        except Exception as e:
            if debug:
                logging.error(
                    "From diff file '%s', %d IDs couldn't be synced because: %s\n%s"
                    % (diff_file, e, len(batchids)))
                pickfile = "batch_sync_updater_%s_%s.pickle" % (
                    bcnt, os.path.basename(diff_file))
                logging.error("IDs pickled in '%s'" % pickfile)
                pickle.dump(batchids, open(pickfile, "wb"))
            raise


class SyncerManager(BaseManager):
    def __init__(self, *args, **kwargs):
        """
        SyncerManager deals with the different syncer objects used to synchronize
        different collections or indices using diff files
        """
        super(SyncerManager, self).__init__(*args, **kwargs)
        self.setup_log()

    def clean_stale_status(self):
        src_build = get_src_build()
        for build in src_build.find():
            dirty = False
            for job in build.get("jobs", []):
                if job.get("status") == "syncing":
                    logging.warning(
                        "Found stale build '%s', marking sync status as 'canceled'"
                        % build["_id"])
                    job["status"] = "canceled"
                    dirty = True
            if dirty:
                src_build.replace_one({"_id": build["_id"]}, build)

    def register_syncer(self, klass):
        if type(klass) == partial:
            assert type(klass.func) == type, "%s is not a class" % klass.func
            diff_type, target_backend_type = klass.func.diff_type, klass.func.target_backend_type
        else:
            diff_type, target_backend_type = klass.diff_type, klass.target_backend_type

        self.register[(diff_type, target_backend_type)] = partial(
            klass,
            log_folder=btconfig.LOG_FOLDER,
            job_manager=self.job_manager)

    def configure(self, klasses=None):
        """
        Register default syncers (if klasses is None) or given klasses.
        klasses is a list of class, or a list of partial'ly initialized classes.
        """
        klasses = klasses or [
            MongoJsonDiffSyncer, ESJsonDiffSyncer,
            MongoJsonDiffSelfContainedSyncer, ESJsonDiffSelfContainedSyncer
        ]
        for klass in klasses:
            self.register_syncer(klass)

    def setup_log(self):
        self.logger, self.logfile = get_logger("syncmanager")

    def __getitem__(self, diff_target):
        """
        Return an instance of a builder for the build named 'build_name'
        Note: each call returns a different instance (factory call behind the scene...)
        """
        # we'll get a partial class but will return an instance
        pclass = BaseManager.__getitem__(self, diff_target)
        return pclass()

    def sync(self,
             backend_type,
             old_db_col_names,
             new_db_col_names,
             diff_folder=None,
             batch_size=100000,
             mode=None,
             target_backend=None,
             steps=["mapping", "content", "meta", "post"],
             debug=False):
        if hasattr(btconfig, "SYNC_BATCH_SIZE"):
            batch_size = btconfig.SYNC_BATCH_SIZE
            self.logger.debug("Overriding sync batch_size default to %s" %
                              batch_size)
        if diff_folder is None:
            diff_folder = generate_folder(btconfig.DIFF_PATH, old_db_col_names,
                                          new_db_col_names)
        if not os.path.exists(diff_folder):
            raise FileNotFoundError(
                "Directory '%s' does not exist, run a diff first" %
                diff_folder)
        # load metadata to know collections that have been diffed in diff_func protocol
        try:
            meta = json.load(open(os.path.join(diff_folder, "metadata.json")))
        except FileNotFoundError:
            self.logger.error("Can't find metadata file in diff folder '%s'" %
                              diff_folder)
            raise

        self.logger.info("Found metadata information: %s" % pformat(meta))
        try:
            diff_type = meta["diff"]["type"]
        except KeyError:
            msg = "Can't find diff_type in metadata file located in '%s'" % diff_folder
            raise SyncerException(msg)

        try:
            syncer = self[(diff_type, backend_type)]
            self.logger.info("Selected syncer: %s" % syncer)
            syncer.old = create_backend(old_db_col_names)
            syncer.new = create_backend(new_db_col_names)
            job = syncer.sync(diff_folder,
                              batch_size=batch_size,
                              mode=mode,
                              target_backend=target_backend,
                              steps=steps,
                              debug=debug)
            return job
        except KeyError as e:
            raise SyncerException("No such syncer (%s,%s) (error: %s)" %
                                  (diff_type, backend_type, e))



'''
Deprecated. This module is not used any more
'''

import types
import logging

enc = None
dec = None
try:
    import bson
    enc = bson.BSON.encode
    dec = bson.BSON.decode
except ImportError:
    import json
    enc = json.dumps
    dec = json.loads

from biothings.utils.common import infer_types


class ParserValidator(object):
    def __init__(self, logger=logging, max_kept_errors=None):
        self.logger = logger
        self.max_kept_errors = max_kept_errors
        self.data_map = {}
        self.summary = {"total": 0}
        self.errs = {}

    def set_err(self, err_type, data):
        self.errs.set_default(err_type, {"count": 0, "data": []})
        self.errs[err_type]["count"] += 1
        if self.max_kept_errors is not None and self.errs[err_type][
                "count"] <= self.max_kept_errors:
            self.errs[err_type]["data"].append(data)

    def analyze(self, data, check_id=True, root_key=None):
        if check_id and "_id" not in data:
            self.set_err("_id_missing", data)
        infer_types(data, self.data_map)

    def check(self, data):
        try:
            data = dec(enc(data))
        except TypeError:
            self.set_err("not_json_serializable", data)
        self.analyze(data)

    def validate(self, func, *args, **kwargs):
        """
        Validate parsing function "func". args and kwargs are
        passed as arguments to this function.
        """
        assert callable(func), "%s is not callable" % func
        res = func(*args, **kwargs)
        assert isinstance(res, list) or isinstance(res, types.GeneratorType), "%s returned a wrong type: %s" % (func, type(res))
        self.logger.info("Return type is %s, now iterating over content" %
                         type(res))
        cnt = 0
        for data in res:
            if not type(data) == dict:
                self.set_err("wrong_data_type", oata)
            self.check(data)
            cnt += 1
            if cnt % 1000 == 0:
                self.logger.debug("Processed %s records" % cnt)



import types
from datetime import datetime
import socket

from biothings.utils.hub_db import get_api
from biothings.utils.loggers import get_logger
from biothings.utils.manager import BaseManager
import biothings.web

from biothings import config as btconfig


class APIManagerException(Exception):
    pass


class APIManager(BaseManager):
    def __init__(self, log_folder=None, *args, **kwargs):
        self.api = get_api()
        self.register = {}
        self.timestamp = datetime.now()
        self.log_folder = log_folder or btconfig.LOG_FOLDER
        self.setup()
        self.restore_running_apis()

    def setup(self):
        self.setup_log()

    def setup_log(self):
        self.logger, _ = get_logger('apimanager')

    def restore_running_apis(self):
        """
        If some APIs were running but the hub stopped, re-start APIs
        as hub restarts
        """
        apis = self.get_apis()
        # these were running but had to stop when hub stopped
        running_apis = [api for api in apis if api.get("status") == "running"]
        for api in running_apis:
            self.logger.info("Restart API '%s'" % api["_id"])
            self.start_api(api["_id"])

    def register_status(self, api_id, status, **extra):
        apidoc = self.api.find_one({"_id": api_id})
        apidoc.update(extra)
        # clean according to status
        if status == "running":
            apidoc.pop("err", None)
        else:
            apidoc.pop("url", None)
        apidoc["status"] = status
        self.api.save(apidoc)

    def get_apis(self):
        return [d for d in self.api.find()]

    def start_api(self, api_id):
        apidoc = self.api.find_one({"_id": api_id})
        if not apidoc:
            raise ApiManagerException("No such API with ID '%s'" % api_id)
        if "entry_point" in apidoc:
            raise NotImplementedError("Custom entry point not implemented yet, "
                                      + "only basic generated APIs are currently supported")
        config_mod = types.ModuleType("config_mod")
        config_str = """
from biothings.web.handlers import BiothingHandler
from biothings.web.handlers import MetadataSourceHandler
from biothings.web.handlers import MetadataFieldHandler
from biothings.web.handlers import QueryHandler

ES_HOST = "%(es_host)s"
ES_INDEX = "%(index)s"
ES_DOC_TYPE = "%(doc_type)s"

# doc_type involved there:
APP_LIST = [
    (r"/metadata/?", MetadataSourceHandler),
    (r"/metadata/fields/?", MetadataFieldHandler),
    (r"/%(doc_type)s/(.+)/?", BiothingHandler),
    (r"/%(doc_type)s/?$", BiothingHandler),
    (r"/query/?", QueryHandler),
]
""" % apidoc["config"]
        code = compile(config_str, "<string>", "exec")
        # propagate config on module
        eval(code, {}, config_mod.__dict__)
        app = biothings.web.BiothingsAPIApp(config_module=config_mod)
        port = int(apidoc["config"]["port"])
        try:
            server = app.get_server()
            server.listen(port)
            self.register[api_id] = server
            self.logger.info("Running API '%s' on port %s" % (api_id, port))
            url = "http://%s:%s" % (socket.gethostname(), port)
            self.register_status(api_id, "running", url=url)
        except Exception as e:
            self.logger.exception("Failed to start API '%s'" % api_id)
            self.register_status(api_id, "failed", err=str(e))
            raise

    def stop_api(self, api_id):
        try:
            assert api_id in self.register, "API '%s' is not running" % api_id
            server = self.register.pop(api_id)
            server.stop()
            if server._stopped:
                self.register_status(api_id, "stopped")
        except Exception as e:
            self.logger.exception("Failed to stop API '%s'" % api_id)
            self.register_status(api_id, "failed", err=str(e))
            raise

    def delete_api(self, api_id):
        try:
            self.stop_api(api_id)
        except Exception as e:
            self.logger.warning("While trying to stop API '%s': %s" %
                                (api_id, e))
        finally:
            self.api.remove({"_id": api_id})

    def create_api(self,
                   api_id,
                   es_host,
                   index,
                   doc_type,
                   port,
                   description=None,
                   **kwargs):
        apidoc = {
            "_id": api_id,
            "config": {
                "port": port,
                "es_host": es_host,
                "index": index,
                "doc_type": doc_type,
            },
            "description": description,
        }
        apidoc.update(kwargs)
        self.api.save(apidoc)
        return apidoc
